<?php
use \mcpay as mcpay;
use PrestaShop\PrestaShop\Core\Payment\PaymentOption;

if (!defined('_PS_VERSION_')) exit;

class Micropayment extends PaymentModule
{
  protected $selectedTab = false;
  var $startURL = 'https://resources.micropayment.de/download/shop-plugins/prestashop/landingpage/';

  public function __construct()
  {
    $this->name = 'micropayment';
    $this->tab = 'payments_gateways';
    $this->version = '1.0.1';
    $this->ps_versions_compliancy = array('min' => '1.7', 'max' => _PS_VERSION_);
    $this->author = 'Micropayment (TE)';
    $this->module_key = '1234deadbeef5678';
    $this->need_instance = 1;

    $this->bootstrap = false;
    parent::__construct();

    if (!defined('MCPAY_BASE_PATH')) {
      define('MCPAY_BASE_PATH', dirname(__FILE__).'/vendor/mcpay/');
    }
    require_once(MCPAY_BASE_PATH.'classes/class.mcpay.php');
    $this->mcpay = new mcpay();

    $this->displayName = 'Micropayment';
    $this->description = 'Accepts payments by Micropayment';
    if ($this->l('BACKEND_TT_DELETE_DETAILS') == "BACKEND_TT_DELETE_DETAILS") {
      $this->confirmUninstall =  "Are you sure you want to delete your details ?";
    } else {
      $this->confirmUninstall = $this->l('BACKEND_TT_DELETE_DETAILS');
    }
  }

  /**
   * install
   *
   * @return bool|string
   */
  public function install()
  {
    $this->warning = null;
    if (is_null($this->warning) && !function_exists('curl_init')) {
      if ($this->l('ERROR_MESSAGE_CURL_REQUIRED') == "ERROR_MESSAGE_CURL_REQUIRED") {
        $this->warning = "cURL is required to use this module. Please install the php extention cURL.";
      } else {
        $this->warning = $this->l('ERROR_MESSAGE_CURL_REQUIRED');
      }
    }
    if (is_null($this->warning)
      && !(parent::install()
        && $this->registerHook('paymentReturn')
        && $this->registerHook('header')
        && $this->registerHook('paymentOptions'))) {
      if ($this->l('ERROR_MESSAGE_INSTALL_MODULE') == "ERROR_MESSAGE_INSTALL_MODULE") {
        $this->warning = "There was an Error installing the module.";
      } else {
        $this->warning = $this->l('ERROR_MESSAGE_INSTALL_MODULE');
      }
    }
    if (is_null($this->warning) && !$this->addMCPOrderStatus()) {
      if ($this->l('ERROR_MESSAGE_CREATE_ORDER_STATUS') == "ERROR_MESSAGE_CREATE_ORDER_STATUS") {
        $this->warning = "There was an Error creating a custom order status.";
      } else {
        $this->warning = $this->l('ERROR_MESSAGE_CREATE_ORDER_STATUS');
      }
    }

    // default setting.
    $confValues = $this->getConfiguration();
    foreach($confValues AS $k => $v) {
      Configuration::updateValue($k, $v);
    }

    $defaultSort = 1;
    foreach ($this->getMethods() AS $pmID => $pmData) {
      //default enable payment grop
      $pmID = strtoupper($pmID);
      Configuration::updateValue('MICROPAYMENT_'.$pmID.'_ACTIVE', '1'); // default paymethod on
      Configuration::updateValue('MICROPAYMENT_'.$pmID.'_TESTMODE', '1'); // default testmode on
      $defaultSort++;
    }

    return is_null($this->warning);
  }

  /**
   * uninstall
   *
   * @return bool
   */
  public function uninstall()
  {
    $generalSettings = true;
    $confValues = $this->getConfiguration();
    foreach($confValues AS $k => $v) {
      if (!$generalSettings) continue; // skip on error
      if(!Configuration::deleteByName($k)){
        $generalSettings = false;
      }
    }

    $paymentSettings = true;
    foreach ($this->getMethods() AS $pmID => $pmData) {
      if (!$paymentSettings) continue; // skip on error
      $pmID = strtoupper($pmID);
      if(!Configuration::deleteByName('MICROPAYMENT_'.$pmID.'_ACTIVE')){
        $paymentSettings = false;
      }
      if(!Configuration::deleteByName('MICROPAYMENT_'.$pmID.'_TESTMODE')){
        $paymentSettings = false;
      }
    }

    if (!$generalSettings || !$paymentSettings
      || !$this->unregisterHook('paymentReturn')
      || !$this->unregisterHook('header')
      || !$this->unregisterHook('paymentOptions')
      || !parent::uninstall()) {
      return false;
    }
    return true;
  }

  /**
   * addMCPOrderStatus
   *
   * @return bool
   */
  public function addMCPOrderStatus()
  {
    $defStateConfig = array(
      'sendEmail' => false,
      'color'     => '#ff0000',
      'hidden'    => false,
      'delivery'  => false,
      'logable'   => false,
      'invoice'   => false,
      'paid'      => false,
      'template'  => '',
    );
    //try {
      $stateConfig = $defStateConfig;
      $stateConfig['color'] = '#FDD835';
      $this->addOrderStatus(
        'MICROPAYMENT_PAYMENT_STATUS_PENDING',
        'MCP wartend',
        $stateConfig
      );

      $stateConfig = $defStateConfig;
      $stateConfig['color'] = '#F44336';
      $stateConfig['sendEmail'] = true;
      $stateConfig['template'] = 'payment_error';
      $this->addOrderStatus(
        'MICROPAYMENT_PAYMENT_STATUS_FAILED',
        'MCP Fehler',
        $stateConfig
      );

      $stateConfig = $defStateConfig;
      $stateConfig['color'] = '#BA68C8';
      $stateConfig['hidden'] = true;
      $this->addOrderStatus(
        'MICROPAYMENT_PAYMENT_STATUS_INVALID',
        'MCP Setupfehler',
        $stateConfig
      );

      $stateConfig = $defStateConfig;
      $stateConfig['color'] = '#26C6DA';
      $stateConfig['sendEmail'] = true;
      $stateConfig['logable'] = true;
      $stateConfig['template'] = 'payment';
      $this->addOrderStatus(
        'MICROPAYMENT_PAYMENT_STATUS_PARTPAID',
        'MCP Teilzahlung',
        $stateConfig
      );

      $stateConfig = $defStateConfig;
      $stateConfig['color'] = '#4CAF50';
      $stateConfig['sendEmail'] = true;
      $stateConfig['logable'] = true;
      $stateConfig['paid'] = true;
      $stateConfig['template'] = 'payment';
      $this->addOrderStatus(
        'MICROPAYMENT_PAYMENT_STATUS_PAID',
        'MCP Bezahlt',
        $stateConfig
      );

      $stateConfig = $defStateConfig;
      $stateConfig['color'] = '#FFB74D';
      $stateConfig['sendEmail'] = true;
      $stateConfig['logable'] = true;
      $stateConfig['template'] = 'refund';
      $this->addOrderStatus(
        'MICROPAYMENT_PAYMENT_STATUS_PARTREFUND',
        'MCP Teilerstattung',
        $stateConfig
      );

      $stateConfig = $defStateConfig;
      $stateConfig['color'] = '#FB8C00';
      $stateConfig['sendEmail'] = true;
      $stateConfig['logable'] = true;
      $stateConfig['paid'] = true;
      $stateConfig['template'] = 'refund';
      $this->addOrderStatus(
        'MICROPAYMENT_PAYMENT_STATUS_REFUND',
        'MCP Erstattung',
        $stateConfig
      );

      return true;
    //} catch (Exception $exception) {
    //  return false;
    //}
  }

  /**
   * addOrderStatus
   *
   * @param $configKey
   * @param $statusName
   * @param $stateConfig
   *
   * @throws \PrestaShopDatabaseException
   * @throws \PrestaShopException
   */
  public function addOrderStatus($configKey, $statusName, $stateConfig)
  {
    if (!Configuration::get($configKey)) {
      $orderState              = new OrderState();
      $orderState->name        = array();
      $orderState->module_name = $this->name;
      $orderState->send_email  = $stateConfig['sendEmail'];
      $orderState->color       = $stateConfig['color'];
      $orderState->hidden      = $stateConfig['hidden'];
      $orderState->delivery    = $stateConfig['delivery'];
      $orderState->logable     = $stateConfig['logable'];
      $orderState->invoice     = $stateConfig['invoice'];
      $orderState->paid        = $stateConfig['paid'];
      foreach (Language::getLanguages() as $language) {
        $orderState->template[$language['id_lang']] = $stateConfig['template'];
        $orderState->name[$language['id_lang']]     = $statusName;
      }

      if ($orderState->add()) {
        $icon = dirname(__FILE__).'/logo.png';
        $newStateIcon = dirname(__FILE__).'/../../img/os/'.(int)$orderState->id.'.gif';
        copy($icon, $newStateIcon);
      }

      Configuration::updateValue($configKey, (int)$orderState->id);
    }
  }

  /**
   * getContent
   *
   * @return string
   */
  public function getContent()
  {
    $output = '';
    //echo '<pre style="padding: 200px">'.print_r($_POST,true).'</pre>';

    // this part is executed only when the form is submitted
    if (Tools::isSubmit('submit' . $this->name)) {
      // retrieve the value set by the user
      //$configValue = (string) Tools::getValue('MYMODULE_CONFIG');
      $conf = $this->loadPluginConfig();
      //echo '<pre>'.print_r($conf,true).'</pre>';
      $fieldsWithBrackets = array('MICROPAYMENT_GENERAL_PAYTEXT', 'MICROPAYMENT_GENERAL_PAYTITLE');

      $error = false;
      foreach($conf AS $k => $v) {
        if ($error) continue;
        // check that the value is valid
        if (!empty($v) && !in_array($k, $fieldsWithBrackets) && !Validate::isGenericName($v)) {
          // invalid value, show an error
          $output = $this->displayError($this->l('Invalid Configuration value')).' '.$k;
          $error = true;
        } else {
          // value is ok, update it and display a confirmation message
          Configuration::updateValue($k, $v);
        }
      }
      if (!$error) {
        $output = $this->displayConfirmation($this->l('Settings updated'));
      }
    }

    // display any message, then the form
    return $output . $this->displayForm();
  }

  /**
   * displayForm
   *
   * @return string
   */
  public function displayForm()
  {
    $shopDomainSsl = Tools::getShopDomainSsl(true, true);
    $backOfficeJsUrl = $shopDomainSsl.__PS_BASE_URI__.'modules/'.$this->name.'/views/js/backoffice.js';
    $backOfficeCssUrl = $shopDomainSsl.__PS_BASE_URI__.'modules/'.$this->name.'/views/css/backoffice.css';

    $tplVars = array(
      'tabs' => $this->getConfigurationTabs(),
      'selectedTab' => $this->getSelectedTab(),
      'backOfficeJsUrl' => $backOfficeJsUrl,
      'backOfficeCssUrl' => $backOfficeCssUrl
    );
    $this->context->smarty->assign($tplVars);

    return $this->display(__FILE__, 'views/templates/admin/tabs.tpl');
  }

  protected function getConfigurationTabs()
  {
    $tabs = array();

    $conf = $this->loadPluginConfig();
    //echo '<pre>'.print_r($conf,true).'</pre>';
    $locale = $this->getLocale();
    // Init Fields form array
    $form = [
      'form' => [
        'legend' => [
          'title' => $this->l('Settings'),
        ],
        'input' => [],
        'submit' => [
          'title' => $this->l('Save'),
          'class' => 'btn btn-default pull-right',
        ],
      ],
    ];
    $form2 = [
      'form' => [
        'legend' => [
          'title' => $this->l('Payment'),
        ],
        'input' => [],
        'submit' => [
          'title' => $this->l('Save'),
          'class' => 'btn btn-default pull-right',
        ],
      ],
    ];

    $helper = new HelperForm();

    // Module, token and currentIndex
    $helper->table = $this->table;
    $helper->name_controller = $this->name;
    $helper->token = Tools::getAdminTokenLite('AdminModules');
    $helper->currentIndex = AdminController::$currentIndex . '&' . http_build_query(['configure' => $this->name]);
    $helper->submit_action = 'submit' . $this->name;

    // Default language
    $helper->default_form_language = (int) Configuration::get('PS_LANG_DEFAULT');

    foreach($conf AS $k => $v){
      if ($k == 'MICROPAYMENT_GENERAL_TESTMODE') continue;
      $type = 'text';
      $formID = 'form';
      if (strpos($k, '_ACTIVE') !== FALSE) {
        $type = 'switch';
        $formID = 'form2';
      }
      if (strpos($k, '_TESTMODE') !== FALSE) {
        $type = 'switch';
        $formID = 'form2';
      }
      if (strpos($k, '_DESCRIPTION') !== FALSE) {
        $type = 'text';
        $formID = 'form2';
      }
      if (strpos($k, '_DEBUG') !== FALSE) {
        $type = 'switch';
      }
      if ($formID == 'form2'){
        $form2['form']['input'][] = [
          'type'     => $type,
          'label'    => $locale[$k],
          'name'     => $k,
          'size'     => 40,
          'required' => TRUE,
        ];
      } else {
        $form['form']['input'][] = [
          'type'     => $type,
          'label'    => $locale[$k],
          'name'     => $k,
          'size'     => 40,
          'required' => TRUE,
        ];
      }
      // Load current value into the form
      $helper->fields_value[$k] = $v;
    }
    //echo '<pre style="padding-left: 200px">'.print_r($conf,true).'</pre>';
    //echo '<pre style="padding-left: 200px">'.print_r($form,true).'</pre>';

    //return $helper->generateForm([$form]).$helper->generateForm([$form2]);

    $path = Media::getMediaPath(_PS_MODULE_DIR_.$this->name.'/');
    try{
      $content = $this->mcpay->doCall($this->startURL, array());
    } catch (Exception $exception) {
      $content = '<img src="'.$path.'mcp_start.jpg">';
    }
    $tplVars = array(
      'startContent' => $content,
      'updateURL' => $this->checkUpdateURL(),
    );
    //echo '<pre style="padding: 200px">'.print_r($tplVars,true).'</pre>';
    $this->context->smarty->assign($tplVars);

    $tabs[] = array(
      'id' => 'presentation',
      'title' => 'Micropayment',
      'content' => $this->display(__FILE__, 'views/templates/admin/start.tpl'),
    );

    $tabs[] = array(
      'id' => 'general_setting',
      'title' => 'Micropayment Setting',
      'content' => $helper->generateForm([$form])
    );

    $tabs[] = array(
      'id' => 'payment_configuration',
      'title' => 'Payment Settings',
      'content' => $helper->generateForm([$form2])
    );

    $tabs[] = array(
      'id' => 'plugin_update_check',
      'title' => 'Update Check',
      'content' => $this->display(__FILE__, 'views/templates/admin/update.tpl'),
    );

    return $tabs;
  }

  public function checkUpdateURL()
  {
    $conf = $this->loadPluginConfig();
    $cfg = $this->getCfg($conf);
    $infos   = array(
      'project'       => $this->mcpay->project,
      'host'          => $_SERVER['HTTP_HOST'],
      'shoptype'      => 'presta',
      'shopversion'   => _PS_VERSION_,
      'plugintype'    => 'WEB+API',
      'pluginversion' => (string)$this->version,
      'status'        => 'stable', // check for stable plugins only (all, stable, beta and alpha)
    );
    $content = $this->mcpay->getVersionURL()."?d=". str_replace('"', "'", serialize($infos));
    return $content;
  }

  protected function getSelectedTab()
  {
    if ($this->selectedTab) {
      return $this->selectedTab;
    }

    if(isset($_POST['MICROPAYMENT_MCP_CCARD_WEB_ACTIVE'])){
      return 'payment_configuration';
    }
    if(isset($_POST['MICROPAYMENT_GENERAL_ACCESSKEY'])){
      return 'general_setting';
    }

    if (Tools::getValue('selected_tab')) {
      return Tools::getValue('selected_tab');
    }

    return 'presentation';
  }

  /**
   * checkCurrency
   *
   * @param $cart
   *
   * @return bool
   */
  public function checkCurrency($cart)
  {
    $currencyOrder = new Currency($cart->id_currency);
    $currencyModules = $this->getCurrency($cart->id_currency);
//    echo '<pre>'.print_r($currencyOrder,true).'</pre>';
//    echo '<pre>'.print_r($currencyModules,true).'</pre>';

    if (is_array($currencyModules)) {
      foreach ($currencyModules as $currencyModule) {
        if ($currencyOrder->id == $currencyModule['id_currency']) {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * hookHeader
   *
   * @param $parameters
   */
  public function hookHeader($parameters)
  {
    //$this->context->controller->addCSS(($this->_path).'views/css/payment_options.css', 'all');
  }

  /**
   * hookPaymentReturn
   *
   * @param $parameters
   */
  public function hookPaymentReturn($parameters)
  {
    if (!$this->active) {
      return;
    }
  }

  /**
   * hookPaymentOptions
   *
   * @param $parameters
   *
   * @return array|void
   */
  public function hookPaymentOptions($parameters)
  {
    if (!$this->active) {
      return;
    }

    if (!$this->checkCurrency($parameters['cart'])) {
      return;
    }

    $conf = $this->loadPluginConfig();
    //echo '<pre>'.print_r($conf,true).'</pre>';
    $cfg = $this->getCfg($conf); // set the config in the mcp object
    //echo '<pre>'.print_r($cfg,true).'</pre>';
    
    $paymentOptions = array();
    foreach ($this->getMethods() AS $pmID => $pmData) {
      $PMID = strtoupper($pmID);

      if (empty($conf['MICROPAYMENT_'.$PMID.'_ACTIVE'])) continue; // hide disabled

      $newOption = new PaymentOption();
      $paymentController = $this->context->link->getModuleLink(
        $this->name,
        $this->mcpay->camelCase('payment_'.$pmID),
        array(),
        true
      );
      $logo = Media::getMediaPath(_PS_MODULE_DIR_.$this->name.'/views/logos/'.$pmData['logo']);

      $paymentName = $pmData['name'];
      $paymentName = '';
      $logoHtml    = '';
      $domain = Tools::getHttpHost(true).__PS_BASE_URI__;

      if ($PMID == 'MCP_CCARD_API'){
        $cfg = array(
          'payformid'     => 'header', // 'conditions-to-approve',
          'accountMode'   => 0, // guest or not
          'jsScriptURL'   => $domain.'module/micropayment/notify?function=getScript&sn=', // https://php74.docker:8081/presta/module/micropayment/notify?function=getScript&sn=mcpay_formstyle_ccard_check
          'jsStyleURL'    => $domain.'module/micropayment/notify?function=getStyle&sn=',
          'fileExtJS'     => '',
        );
        $data = array(
          'classes'       => array(),
          'errorMSG'      => '',
          'mcpay_fields'  => array(),
        );
        // translations to the default texts in template file. Search and replace is done.
        $translation = $this->getCCardLocale();
        $tpl = $this->mcpay->initCCardIFrame($cfg, $data, $translation);
        $logoHtml .= $tpl;
      } else if ($PMID == 'MCPAY_PREPAY_API'){

      } else if ($PMID == 'MCPAY_SEPA_API'){

      } else { // web payments
        if (!empty($conf['MICROPAYMENT_'.$PMID.'_DESCRIPTION'])){
          $logoHtml .= '<span>'.$conf['MICROPAYMENT_'.$PMID.'_DESCRIPTION'].'</span>';
        } else {
          $logoHtml .= '<b>'.$pmData['name'].'</b>';
        }
      }
      $logoHtml .= $this->mcpay->getPoweredBy();
      $logoHtml .= '<br><br>';

      /*
      $callToActionText: The text displayed as the payment method name.
      $additionalInformation: Additional information to display to the customer. This is free HTML, and may be used by modules such as bankwire to display to which account the bank transfer should be made.
      $logo: The URL to a picture to display in the payment option selection widget.
      $action: The URL to which the request to process the payment must be made.
      $inputs: An associative array of additional parameters to use when sending the request to $action.
      $form: The custom HTML to display like a form to enter the credit card information.
      $iframe: The custom HTML containing an iframe with the payment confirmation widget for modules like ATOS.
      $moduleName: The name of the module.
      $binary: A boolean to set if the module form was generated by binaries and contains a submit button. It’s necessary to adapt the behavior.
      */

      $newOption->setCallToActionText($paymentName)
        ->setAction($paymentController)
        ->setAdditionalInformation($logoHtml)
        //->setForm('<b>blabla</b>')
        ->setLogo($logo);

      $paymentOptions[] = $newOption;
    }

    return $paymentOptions;
  }

  /**
   * loadPluginConfig
   *
   * @return array
   */
  public function loadPluginConfig()
  {
    $config = array();
    $confValues = $this->getConfiguration();
    foreach($confValues AS $k => $v) {
      $config[$k] = Tools::getValue($k, Configuration::get($k));
    }

    foreach ($this->getMethods() AS $pmID => $pmData) {
      $pmID = strtoupper($pmID);
      $k = 'MICROPAYMENT_'.$pmID.'_ACTIVE';
      $config[$k] = Tools::getValue($k, Configuration::get($k));
      $k = 'MICROPAYMENT_'.$pmID.'_TESTMODE';
      $config[$k] = Tools::getValue($k, Configuration::get($k));
      $k = 'MICROPAYMENT_'.$pmID.'_DESCRIPTION';
      $config[$k] = Tools::getValue($k, Configuration::get($k));
    }
    return $config;
  }

  /**
   * getMethods
   *
   * @return \string[][]
   */
  private function getMethods()
  {
    $pms = array(
      'mcp_ccard_web' => array(
        'name' => 'Kreditkarte Micropayment',
        'desc' => 'powered by Micropayment',
        'logo' => 'ccard.png',
      ),
      'mcp_sepa_web' => array(
        'name' => 'Lastschrift Micropayment',
        'desc' => 'powered by Micropayment',
        'logo' => 'sepa.png',
      ),
      'mcp_sofort_web' => array(
        'name' => 'Sofort Micropayment',
        'desc' => 'powered by Micropayment',
        'logo' => 'sofort.png',
      ),
      'mcp_paypal_web' => array(
        'name' => 'PayPal Micropayment',
        'desc' => 'powered by Micropayment',
        'logo' => 'paypal.png',
      ),
      'mcp_prepay_web' => array(
        'name' => 'Vorkasse Micropayment',
        'desc' => 'powered by Micropayment',
        'logo' => 'prepay_de.png',
      ),
      'mcp_paysafecard_web' => array(
        'name' => 'paysafecard Micropayment',
        'desc' => 'powered by Micropayment',
        'logo' => 'paysafecard.png',
      ),
      'mcp_paydirekt_web' => array(
        'name' => 'paydirekt Micropayment',
        'desc' => 'powered by Micropayment',
        'logo' => 'paydirekt.png',
      ),
      // API Paymethods
//      'mcp_ccard_api' => array(
//        'name' => 'Kreditkarte',
//        'desc' => 'powered by Micropayment',
//        'logo' => 'ccard.png',
//      ),
//      'mcp_sepa_api' => array(
//        'name' => 'Lastschrift',
//        'desc' => 'powered by Micropayment',
//        'logo' => 'sepa.png',
//      ),
//      'mcp_prepay_api' => array(
//        'name' => 'Vorkasse',
//        'desc' => 'powered by Micropayment',
//        'logo' => 'prepay_de.png',
//      ),
    );
    return $pms;
  }

  /**
   * getConfiguration
   *
   * @return string[]
   */
  private function getConfiguration()
  {
    $config = array(
      'MICROPAYMENT_GENERAL_ACCESSKEY'   => '', //'9e7774d41a312c464cd4b8b36b7954a2',
      'MICROPAYMENT_GENERAL_BGCOLOR'     => '',
      'MICROPAYMENT_GENERAL_BGGFX'       => '',
      'MICROPAYMENT_GENERAL_DEBUG'       => '1',
      'MICROPAYMENT_GENERAL_SUFFIX'      => 'micropayment.de',
      'MICROPAYMENT_GENERAL_GFX'         => '',
      'MICROPAYMENT_GENERAL_PROJECTID'   => '', //'11cu-prast-112ad16f', //'11cu-jovet-0df2fd78',
      'MICROPAYMENT_GENERAL_SECRETFIELD' => '', //'351930fc02dfc484b36601d4de63aec374e647a9',
      'MICROPAYMENT_GENERAL_TESTMODE'    => '1',
      'MICROPAYMENT_GENERAL_THEME'       => 'x1',
      'MICROPAYMENT_GENERAL_EXPIREDAYS'  => '21',
      'MICROPAYMENT_GENERAL_PAYTEXT'     => 'Order {$orderid}',
      'MICROPAYMENT_GENERAL_PAYTITLE'    => 'Order {$orderid}',
    );
    return $config;
  }

  /**
   * getCfg
   *
   * @param $config
   *
   * @return array
   */
  public function getCfg($config = array())
  {
    if (empty($config)){
      $config = $this->loadPluginConfig();
      //echo '<pre>'.print_r($config,true).'</pre>';
    }
    // key and value tauschen
    $configMap = array(
      'MICROPAYMENT_GENERAL_ACCESSKEY'    => 'accessKey',
      'MICROPAYMENT_GENERAL_BGCOLOR'      => 'bgcolor',
      'MICROPAYMENT_GENERAL_BGGFX'        => 'bggfx',
      'MICROPAYMENT_GENERAL_DEBUG'        => 'debug',
      'MICROPAYMENT_GENERAL_SUFFIX'       => 'suffix',
      'MICROPAYMENT_GENERAL_GFX'          => 'gfx',
      'MICROPAYMENT_GENERAL_PROJECTID'    => 'project',
      'MICROPAYMENT_GENERAL_SECRETFIELD'  => 'secretfield',
      'MICROPAYMENT_GENERAL_TESTMODE'     => 'testMode',
      'MICROPAYMENT_GENERAL_THEME'        => 'theme',
      'MICROPAYMENT_GENERAL_EXPIREDAYS'   => 'expiredays',
      'MICROPAYMENT_GENERAL_PAYTEXT'      => 'paytext',
      'MICROPAYMENT_GENERAL_PAYTITLE'     => 'paytitle',
    );
    $cfg       = array();
    foreach ($configMap as $k => $v) {
      if (empty($v)) continue;
      if (empty($config[$k])) {
        $cfg[$v] = '';
      } else {
        $cfg[$v] = $config[$k];
      }
    }
    //$cfg['paytext'] = 'Your Order: {$orderid}';
    //mail('webmaster@web-dezign.de', __CLASS__.'->'.__FUNCTION__.'->$cfg', print_r($cfg,1).print_r($config,1));
    //echo '<pre>'.print_r($cfg,true).'</pre>';
    $this->mcpay->setConfig($cfg);

    return $cfg;
  }

  /**
   * getLocale
   *
   * @return array
   */
  private function getLocale()
  {
    $locale = array();

    $key = 'MICROPAYMENT_GENERAL_ACCESSKEY';
    $locale[$key] = $this->l('MICROPAYMENT_GENERAL_ACCESSKEY');
    if ($locale[$key] == $key) $locale[$key] = 'AccessKey';

    $key = 'MICROPAYMENT_GENERAL_BGCOLOR';
    $locale[$key] = $this->l('MICROPAYMENT_GENERAL_BGCOLOR');
    if ($locale[$key] == $key) $locale[$key] = 'Hintergrundfarbe';

    $key = 'MICROPAYMENT_GENERAL_BGGFX';
    $locale[$key] = $this->l('MICROPAYMENT_GENERAL_BGGFX');
    if ($locale[$key] == $key) $locale[$key] = 'Hintergrundbild';

    $key = 'MICROPAYMENT_GENERAL_DEBUG';
    $locale[$key] = $this->l('MICROPAYMENT_GENERAL_DEBUG');
    if ($locale[$key] == $key) $locale[$key] = 'Debugmode';

    $key = 'MICROPAYMENT_GENERAL_SUFFIX';
    $locale[$key] = $this->l('MICROPAYMENT_GENERAL_SUFFIX');
    if ($locale[$key] == $key) $locale[$key] = 'Micropayment Domain';

    $key = 'MICROPAYMENT_GENERAL_GFX';
    $locale[$key] = $this->l('MICROPAYMENT_GENERAL_GFX');
    if ($locale[$key] == $key) $locale[$key] = 'Logobild';

    $key = 'MICROPAYMENT_GENERAL_PROJECTID';
    $locale[$key] = $this->l('MICROPAYMENT_GENERAL_PROJECTID');
    if ($locale[$key] == $key) $locale[$key] = 'Projektkürzel';

    $key = 'MICROPAYMENT_GENERAL_SECRETFIELD';
    $locale[$key] = $this->l('MICROPAYMENT_GENERAL_SECRETFIELD');
    if ($locale[$key] == $key) $locale[$key] = 'Sicherheitsfeld';

    $key = 'MICROPAYMENT_GENERAL_THEME';
    $locale[$key] = $this->l('MICROPAYMENT_GENERAL_THEME');
    if ($locale[$key] == $key) $locale[$key] = 'Theme';

    $key = 'MICROPAYMENT_GENERAL_EXPIREDAYS';
    $locale[$key] = $this->l('MICROPAYMENT_GENERAL_EXPIREDAYS');
    if ($locale[$key] == $key) $locale[$key] = 'Zahlungsziel (Vorkasse)';

    $key = 'MICROPAYMENT_GENERAL_PAYTEXT';
    $locale[$key] = $this->l('MICROPAYMENT_GENERAL_PAYTEXT');
    if ($locale[$key] == $key) $locale[$key] = 'Freitext';

    $key = 'MICROPAYMENT_GENERAL_PAYTITLE';
    $locale[$key] = $this->l('MICROPAYMENT_GENERAL_PAYTITLE');
    if ($locale[$key] == $key) $locale[$key] = 'Freitext Titel';

    $key = 'MICROPAYMENT_MCP_PREPAY_API_TESTMODE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PREPAY_API_TESTMODE');
    if ($locale[$key] == $key) $locale[$key] = 'Vorkasse API TestMode';

    $key = 'MICROPAYMENT_MCP_PREPAY_API_DESCRIPTION';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PREPAY_API_DESCRIPTION');
    if ($locale[$key] == $key) $locale[$key] = 'Vorkasse API Beschreibung';

    $key = 'MICROPAYMENT_MCP_PREPAY_API_ACTIVE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PREPAY_API_ACTIVE');
    if ($locale[$key] == $key) $locale[$key] = 'Vorkasse API aktivieren';

    $key = 'MICROPAYMENT_MCP_SEPA_API_TESTMODE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_SEPA_API_TESTMODE');
    if ($locale[$key] == $key) $locale[$key] = 'SEPA Lastschrift API TestMode';

    $key = 'MICROPAYMENT_MCP_SEPA_API_DESCRIPTION';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_SEPA_API_DESCRIPTION');
    if ($locale[$key] == $key) $locale[$key] = 'SEPA Lastschrift API Beschreibung';

    $key = 'MICROPAYMENT_MCP_SEPA_API_ACTIVE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_SEPA_API_ACTIVE');
    if ($locale[$key] == $key) $locale[$key] = 'SEPA Lastschrift API aktivieren';

    $key = 'MICROPAYMENT_MCP_CCARD_API_TESTMODE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_CCARD_API_TESTMODE');
    if ($locale[$key] == $key) $locale[$key] = 'Kreditkarte API TestMode';

    $key = 'MICROPAYMENT_MCP_CCARD_API_DESCRIPTION';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_CCARD_API_DESCRIPTION');
    if ($locale[$key] == $key) $locale[$key] = 'Kreditkarte API Beschreibung';

    $key = 'MICROPAYMENT_MCP_CCARD_API_ACTIVE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_CCARD_API_ACTIVE');
    if ($locale[$key] == $key) $locale[$key] = 'Kreditkarte API aktivieren';

    $key = 'MICROPAYMENT_MCP_PAYDIREKT_WEB_TESTMODE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PAYDIREKT_WEB_TESTMODE');
    if ($locale[$key] == $key) $locale[$key] = 'Giropay TestMode';

    $key = 'MICROPAYMENT_MCP_PAYDIREKT_WEB_DESCRIPTION';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PAYDIREKT_WEB_DESCRIPTION');
    if ($locale[$key] == $key) $locale[$key] = 'Giropay Beschreibung';

    $key = 'MICROPAYMENT_MCP_PAYDIREKT_WEB_ACTIVE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PAYDIREKT_WEB_ACTIVE');
    if ($locale[$key] == $key) $locale[$key] = 'Giropay aktivieren';

    $key = 'MICROPAYMENT_MCP_PAYSAFECARD_WEB_TESTMODE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PAYSAFECARD_WEB_TESTMODE');
    if ($locale[$key] == $key) $locale[$key] = 'Paysafecard TestMode';

    $key = 'MICROPAYMENT_MCP_PAYSAFECARD_WEB_DESCRIPTION';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PAYSAFECARD_WEB_DESCRIPTION');
    if ($locale[$key] == $key) $locale[$key] = 'Paysafecard Beschreibung';

    $key = 'MICROPAYMENT_MCP_PAYSAFECARD_WEB_ACTIVE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PAYSAFECARD_WEB_ACTIVE');
    if ($locale[$key] == $key) $locale[$key] = 'Paysafecard aktivieren';

    $key = 'MICROPAYMENT_MCP_PREPAY_WEB_TESTMODE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PREPAY_WEB_TESTMODE');
    if ($locale[$key] == $key) $locale[$key] = 'Vorkasse TestMode';

    $key = 'MICROPAYMENT_MCP_PREPAY_WEB_DESCRIPTION';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PREPAY_WEB_DESCRIPTION');
    if ($locale[$key] == $key) $locale[$key] = 'Vorkasse Beschreibung';

    $key = 'MICROPAYMENT_MCP_PREPAY_WEB_ACTIVE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PREPAY_WEB_ACTIVE');
    if ($locale[$key] == $key) $locale[$key] = 'Vorkasse aktivieren';

    $key = 'MICROPAYMENT_MCP_PAYPAL_WEB_TESTMODE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PAYPAL_WEB_TESTMODE');
    if ($locale[$key] == $key) $locale[$key] = 'PayPal TestMode';

    $key = 'MICROPAYMENT_MCP_PAYPAL_WEB_DESCRIPTION';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PAYPAL_WEB_DESCRIPTION');
    if ($locale[$key] == $key) $locale[$key] = 'PayPal Beschreibung';

    $key = 'MICROPAYMENT_MCP_PAYPAL_WEB_ACTIVE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_PAYPAL_WEB_ACTIVE');
    if ($locale[$key] == $key) $locale[$key] = 'PayPal aktivieren';

    $key = 'MICROPAYMENT_MCP_SOFORT_WEB_TESTMODE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_SOFORT_WEB_TESTMODE');
    if ($locale[$key] == $key) $locale[$key] = 'Sofort TestMode';

    $key = 'MICROPAYMENT_MCP_SOFORT_WEB_DESCRIPTION';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_SOFORT_WEB_DESCRIPTION');
    if ($locale[$key] == $key) $locale[$key] = 'Sofort Beschreibung';

    $key = 'MICROPAYMENT_MCP_SOFORT_WEB_ACTIVE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_SOFORT_WEB_ACTIVE');
    if ($locale[$key] == $key) $locale[$key] = 'Sofort aktivieren';

    $key = 'MICROPAYMENT_MCP_SEPA_WEB_TESTMODE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_SEPA_WEB_TESTMODE');
    if ($locale[$key] == $key) $locale[$key] = 'SEPA Lastschrift TestMode';

    $key = 'MICROPAYMENT_MCP_SEPA_WEB_DESCRIPTION';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_SEPA_WEB_DESCRIPTION');
    if ($locale[$key] == $key) $locale[$key] = 'SEPA Lastschrift Beschreibung';

    $key = 'MICROPAYMENT_MCP_SEPA_WEB_ACTIVE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_SEPA_WEB_ACTIVE');
    if ($locale[$key] == $key) $locale[$key] = 'SEPA Lastschrift aktivieren';

    $key = 'MICROPAYMENT_MCP_CCARD_WEB_TESTMODE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_CCARD_WEB_TESTMODE');
    if ($locale[$key] == $key) $locale[$key] = 'Kreditkarte TestMode';

    $key = 'MICROPAYMENT_MCP_CCARD_WEB_DESCRIPTION';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_CCARD_WEB_DESCRIPTION');
    if ($locale[$key] == $key) $locale[$key] = 'Kreditkarte Beschreibung';

    $key = 'MICROPAYMENT_MCP_CCARD_WEB_ACTIVE';
    $locale[$key] = $this->l('MICROPAYMENT_MCP_CCARD_WEB_ACTIVE');
    if ($locale[$key] == $key) $locale[$key] = 'Kreditkarte aktivieren';

    return $locale;
  }

  /**
   * getCCardLocale
   *
   * @return array
   */
  private function getCCardLocale()
  {
    $locale = array(
      'Name on Card'              => $this->l( 'Card Holder'),
      'Remember Data'             => $this->l( 'Remember Card for future purchase'),
      'Use new Card'              => $this->l( 'Use new Card'),
      'Use old Card'              => $this->l( 'Use old Card'),
      'Please check your inputs!' => $this->l( 'Please check your inputs!'),
      'Vorname Nachname'          => $this->l( 'Firstname Lastname'),
      'Credit Card Number'        => $this->l( 'Card Number'),
      'Card Number'               => $this->l( 'Card Number'),
      'Expiration Date'           => $this->l( 'ExpDate'),
      'ExpDate'                   => $this->l( 'ExpDate'),
      'Card CVC'                  => $this->l( 'CVC'),
      'Month'                     => $this->l( 'Month'),
      'Year'                      => $this->l( 'Year'),
      'NAME'                      => $this->l( 'Card Holder'),
      'PAN'                       => $this->l( 'Card Number'),
      'EXPIRE'                    => $this->l( 'ExpDate'),
    );
    return $locale;
  }

  /**
   * getUserData
   *
   * @param \Shopware\Core\Checkout\Order\OrderEntity $order
   * @param $config
   * @param \Shopware\Core\System\SalesChannel\SalesChannelContext $salesChannelContext
   *
   * @return array
   */
  public function getUserData()
  {
    $cart        = $this->context->cart;
    //$contextLink = $this->context->link;
    //$customer    = new Customer($cart->id_customer);
    $address     = new Address((int)$cart->id_address_delivery);
    $country     = new Country($address->id_country);
    $currency    = new Currency((int)$cart->id_currency);
    //$cartDetails = $cart->getSummaryDetails();

    $orderId = $cart->id;
    $custID = $cart->id_customer;

    $amount = $cart->getOrderTotal(true, Cart::BOTH);
    $countryISO = $country->iso_code;
    $currency = $currency->iso_code;

//        $infos       = $mpBootstrap->getInfo();
    $chksumData = array(
      'amount'        => $amount * 100,
      'currency'      => $currency,
      'iso'           => $countryISO,
      'host'          => $_SERVER['HTTP_HOST'],
      'project'       => $this->project,
      'shoptype'      => 'presta',
      'shopversion'   => _PS_VERSION_,
      'plugintype'    => 'WEB+API',
      'pluginversion' => '1.0.0',//(string)$infos['version'],
      'pm'            => $this->paymentMethod,
    );

    // take browser language
    $lang = strtolower(substr($_SERVER['HTTP_ACCEPT_LANGUAGE'],0,2));
    if(!in_array($lang, array('de', 'en'))){
      $lang = 'de';
    }

    $userData = array(
      'company'    => $this->context->customer->company,
      'firstName'  => $this->context->customer->firstname,
      'lastName'   => $this->context->customer->lastname,
      'salutation' => $this->context->customer->id_gender==0?'mr':'mrs',
      'street'     => $address->address1,
      'zip'        => $address->postcode,
      'city'       => $address->city,
      'country'    => $countryISO,
      'email'      => $this->context->customer->email,
      'ip'         => $_SERVER['REMOTE_ADDR'],
      'id'         => $custID,
      'order_id'   => $orderId,
      'lang'       => $lang, //$order->getLanguage()->getName(), // sadly always empty
      'amount'     => ($amount * 100),
      'currency'   => $currency,
      'chksum'     => $chksumData, // request security
      'doFinish'   => false,
      'guest'      => (bool)($this->context->customer->is_guest == 1),
    );
    if (empty($userData['ip'])) $userData['ip'] = $_SERVER['REMOTE_ADDR']; // Falls IP Leer, dann aus dem Server holen

    return $userData;
  }

  /**
   * getURL
   *
   * @param $userData
   * @param $config
   * @param $activePayment
   *
   * @return string
   * @throws \Exception
   */
  public function getURL($userData, $config, $activePayment)
  {
    if (strpos($activePayment, '_') !== false) {
      $p = explode('_', $activePayment);
      $activePayment = $p[1];
    }
    $config['extraParams']['pm'] = $activePayment;
    $forwardURL = $this->mcpay->getPaymentWindowParams($userData, $config, $activePayment);
    return $forwardURL;
  }
}
?>