<?php
if (!class_exists('mcpay')) {
  if (!defined('MCPAY_BASE_PATH')) {
    $mcpayPath = dirname(__FILE__).'/../';
    define('MCPAY_BASE_PATH', $mcpayPath);
  }
  if (is_dir(MCPAY_BASE_PATH.'classes/')) {
    require_once(MCPAY_BASE_PATH.'classes/class.mcpay_core.php');
  } else if (is_dir(MCPAY_BASE_PATH.'Classes/')) {
    require_once(MCPAY_BASE_PATH.'Classes/class.mcpay_core.php');
  } else {
    die('class.mcpay_core.php not found!');
  }

  class mcpay extends mcpay_core
  {
    var $version = '1.4.0';

    /**
     * mcpay constructor.
     *
     * @param int  $testMode
     * @param null $debugLog
     */
    function __construct($testMode = 1, $debugLog = NULL)
    {
      parent::__construct($testMode, $debugLog);
    }

    /**
     * getPoweredBy
     *
     * @param bool  $js
     * @param array $chksumDat
     * @param int   $status
     * @param null  $comment
     * @param bool  $onlyImg
     *
     * @return string
     */
    public function getPoweredBy($js = FALSE, $chksumDat = array(), $status = 0, $comment = NULL, $onlyImg = FALSE)
    {
      if (!is_array($chksumDat)) {
        $chksumDat = unserialize($chksumDat);
      }
      if (empty($chksumDat['project'])) $chksumDat['project'] = $this->project;
      if (empty($chksumDat['ip'])) $chksumDat['ip'] = $_SERVER['REMOTE_ADDR'];
      if (empty($chksumDat['host'])) $chksumDat['host'] = $_SERVER['HTTP_HOST'];
      if (!empty($comment)) $chksumDat['comment'] = $comment;
      $chksum = serialize($chksumDat);

      if ($onlyImg) {
        return '<img src="'.self::getLogoPicURL().'?d='.base64_encode($chksum).'&s='.$status.'">';
      }

      if ($js) {
        $info = '<script src=\''.self::getLogoURL().'?d='.$chksum.'&s='.$status.'\'></script>';
      } else {
        $info = '<div style="font-size: 10px;display: inline-block;color: #bbb; float: right">'
          .'powered by '
          .'<a href="http://www.micropayment.de" title="Payment Service Provider" target="_blank" style="float: none;margin: 0;color: #000">'
          .'micro<span style="color: #800">payment</span>'
          .'</a>';
        if (!empty($status)) {
          $info .= '<img src="'.self::getLogoPicURL().'?d='.base64_encode($chksum).'&s='.$status.'">';
        }
        $info .= '</div>';
      }
      return $info;
    }

    /**
     * getLogo
     *
     * @param      $paytype
     * @param      $title
     * @param bool $asTag
     *
     * @return string
     */
    public function getLogo($paytype, $title = NULL, $asTag = TRUE)
    {
      $paytype = strtolower($paytype);
      if (!array_key_exists($paytype, $this->iconURLs)) {
        $img = 'https://www.micropayment.de/resources/?what=img&group=logo&show=type4.95x33.white2';
      } else {
        $img = $this->iconURLs[$paytype];
      }
      if (!$asTag) return $img;
      return '<img src="'.$img.'"  title="'.$title.'" style="vertical-align: top; display: inline-block; float: none; padding: 0 10px;"/>';
    }

    /**
     * getAllowedCurrenies
     *
     * @param $config
     *
     * @return mixed
     * @throws \Exception
     */
    function getAllowedCurrenies($config)
    {
      $this->setConfig($config);
      try {
        $res = $this->currenciesGetSupported();
      } catch (Exception $e) {
        throw new Exception('currenciesGetSupported failed. '.$e->getMessage());
      }
      //mail('te@st.de', __CLASS__.'->'.__FUNCTION__, print_r($res,true));
      return $res->result;
    }

    /**
     * bookCCard
     *
     * @param $userData
     * @param $config
     * @param $payData
     *
     * @return bool|\stdClass
     * @throws \Exception
     */
    function bookCCard($userData, $config, $payData)
    {
      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->userData: '.print_r($userData, TRUE));
      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->payData: '.print_r($payData, TRUE));

      if (empty($payData['token'])) {
        throw new Exception(__CLASS__.'->'.__FUNCTION__.': Token cant be empty!');
        //$payData['reuse'] = TRUE;$config
      }

      if (!empty($payData['remember'])) {
        $this->bakePayCookieCC($payData, $userData['id']);
      }

      $retObj = new stdClass();
      $this->setConfig($config);
      $this->switchPaymethod('ccard');
      $token = $payData['token'];

      $custID = $userData['id'];
      if (empty($custID)) $custID = 'guest-'.$userData['order_id']; // guest order
      $custID             = $this->unifyID($custID);
      $custObj            = new stdClass();
      $custObj->id        = $custID;
      $custObj->firstname = $userData['firstName'];
      $custObj->lastname  = $userData['lastName'];
      $custObj->email     = $userData['email'];
      $custObj->street    = $userData['street'];
      $custObj->zip       = $userData['zip'];
      $custObj->city      = $userData['city'];
      $custObj->country   = $userData['country'];
      $custObj->company   = $userData['company'];
      $custObj->lang      = strtolower($userData['lang']); //'de-DE';
      if (!empty($userData['ip'])) {
        $custObj->ip = $userData['ip'];
      } else {
        $custObj->ip = $_SERVER['REMOTE_ADDR'];
      }
      $orderID           = $this->unifyID($userData['order_id']);
      $config['paytext'] = str_replace('{$orderid}', $orderID, $config['paytext']);
      $bookObj           = new stdClass();
      $bookObj->amount   = $userData['amount'];
      $bookObj->currency = $userData['currency'];
      $bookObj->orderID  = $orderID;
      $bookObj->title    = 'Order '.$orderID; // shown in control center
      $bookObj->paytext  = $config['paytext']; // shown on customer bank account
      $bookObj->payload  = array(
        'orderid'    => $orderID,
        'customerid' => $custID,
        'customerip' => $_SERVER['REMOTE_ADDR'],
      );
      if (!empty($config['payload'])) {
        foreach ($config['payload'] as $k => $v) {
          $bookObj->payload[$k] = $v;
        }
      }
      $custObj->payload = $bookObj->payload;
      if (isset($config['autogenSession'])) {
        $bookObj->autogenSession = (bool)$config['autogenSession'];
      }

      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->custObj: '.print_r($custObj, TRUE));
      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->bookObj: '.print_r($bookObj, TRUE));
      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->config: '.print_r($config, TRUE));
      $retObj->custObj = $custObj;
      $retObj->bookObj = $bookObj;

      $checkCC        = TRUE;
      $createCustomer = TRUE;
      if (!empty($payData['forceNewCard'])) {
        $payData['reuse'] = FALSE;
        $checkCC          = FALSE;
      }
      try {
        // customerGet check if customer exists
        $this->customerGet($custObj->id); // check with prefix
        $createCustomer = FALSE; // customer exists
      } catch (Exception $e) {
        // if customer doesnt exists
        if ($e->getCode() == '3110') {
          $checkCC = FALSE; // no customer, no ccard data
          if (strpos($custObj->id, '_') !== FALSE) {
            try {
              $t = explode('_', $custObj->id);
              $this->customerGet($t[1]);
              $custObj->id    = $t[1]; // without prefix
              $checkCC        = TRUE;
              $createCustomer = FALSE; // customer without prefix exists
            } catch (Exception $e) {
              // this customer also doesnt exist
              $createCustomer = TRUE; // customer with and without prefix doesnt exist
            }
          } else {
            // customer id has no prefix in first try and customer dows not exist
            $createCustomer = TRUE;
          }

        } else {
          throw new Exception('customerGet failed. '.$e->getMessage());
        }
      }

      if ($createCustomer) {
        // try to create customer
        try {
          $this->customerCreate($custObj);
        } catch (Exception $e) {
          throw new Exception('customerCreate failed. '.$e->getMessage());
        }
      }

      if ($checkCC) {
        try {
          // check if creditcard data exists
          $res = $this->creditcardDataGet($custObj->id);
          //mail('webmaster@web-dezign.de', __CLASS__.'->'.__FUNCTION__.'->$res', print_r($res,true));
          if (!isset($payData['reuse'])) $payData['reuse'] = TRUE; // if customer has card, reuse it
          if (!empty($res->cvc2Required)) $payData['reuse'] = FALSE; // if cvc is needed card cant be reused
          if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->creditcardDataGet: '.print_r($res, TRUE).print_r($payData, TRUE)); // log result
        } catch (Exception $e) {
          if ($e->getCode() == '3500') {
            $payData['reuse'] = FALSE;
          } else {
            throw new Exception('creditcardDataGet failed. '.$e->getMessage());
          }
        }
      } else {
        $payData['reuse'] = FALSE; // new customer cant reuse
      }

      // if last try was an 3D error, we need to start 3D again
      if (!empty($payData['3Derror'])) {
        $payData['reuse'] = FALSE;
      }

      // create customer and session in one step but with update = 2 for customer update
      if ($payData['reuse']) {
        try {
          $retry = FALSE;
          $res   = $this->sessionCreate($custObj, $bookObj);
          if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->sessionCreate: '.print_r($res, TRUE)); // log result
        } catch (Exception $e) {
          if ($e->getCode() == '3500') { // No credit card data available for customerId
            $retry = TRUE;
          } else if ($e->getCode() == '3110') { // Unknown Customer ID cant be used for reuse!
            $retry = TRUE;
          } else {
            throw new Exception('sessionCreate failed. '.$e->getMessage());
          }
        }
        // handle older subscriptions with old customer ids
        if ($retry) {
          try {
            $t           = explode('_', $custObj->id);
            $custObj->id = $t[1];
            $res         = $this->sessionCreate($custObj, $bookObj);
            if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->sessionCreate-retry: '.print_r($res, TRUE)); // log result
          } catch (Exception $e) {
            throw new Exception('sessionCreate failed. '.$e->getMessage());
          }
        }
      } else {
        try {
          $res = $this->tokenSessionCreate($token, $custObj, $bookObj);
          if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->tokenSessionCreate: '.print_r($res, TRUE)); // log result
        } catch (Exception $e) {
          throw new Exception('tokenSessionCreate failed. '.$e->getMessage());
        }
      }

      try {
        $address = $custObj->street.$custObj->zip.$custObj->city;
        if (!empty($address)) {
          $res2 = $this->addressSet($custObj->id, $custObj);
          if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->addressSet: '.print_r($res, 1)); // log result
        }
      } catch (Exception $e) {
        throw new Exception('addressSet failed. '.$e->getMessage());
      }

      if ($res->error) {
        if ($payData['reuse']) {
          throw new Exception('sessionCreate failed. '.$res->errorMessage);
        } else {
          throw new Exception('tokenSessionCreate failed. '.$res->errorMessage);
        }
      }
      try {
        $testMode       = (int)json_decode($config['testMode']);
        $fraudDetection = 1;
        if (!empty($testMode)) {
          $fraudDetection = 0;
        }
        $avsCheck  = 0;
        $sessionID = $res->sessionId;
        $returnURL = $config['returnURL'];
        if ($payData['reuse'] && empty($payData['force3d'])) { // recurring booking always without 3D but recurring 1
          $start3d   = 0;
          $recurring = 1;
        } else { // initial booking always with 3D
          $start3d   = 1;
          $recurring = 0;
        }
        if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->transactionPurchaseParams: '.' SID:'.$sessionID.' FD:'.$fraudDetection.' AVS:'.$avsCheck.' RECUR:'.$recurring.' RURL:'.$returnURL.' 3D:'.$start3d); // log result
        $res = $this->transactionPurchase($sessionID, $fraudDetection, $avsCheck, $recurring, $returnURL, $start3d);
        if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->transactionPurchase: '.print_r($res, TRUE)); // log result
        $retObj->trxInfo = $res;
        $retObj          = $this->finishCheck($retObj, $userData);

        switch ($res->secureStatus) {
          case 'BILLING':
            if ($res->transactionStatus == 'SUCCESS') {
              $retObj->trxInfo->sessionId = $sessionID; // to have it in return value
              if (!empty($config['returnURLNO3D'])) {
                $retObj->forward    = 1;
                $retObj->forwardURL = $config['returnURLNO3D'].$sessionID;
              }
            } else {
              throw new Exception('Error: '.$res->transactionResultMessage);
            }
            break;
          case 'START3D':
            $retObj->forward    = 1;
            $retObj->forwardURL = $res->forwardUrl;
            break;
          case 'DECLINED':
            throw new Exception('Transaction delined.');
            break;
          case 'FAILURE':
            throw new Exception('Transaction failed.');
            break;
        }
      } catch (Exception $e) {
        $retObj->error        = 1;
        $retObj->errorMessage = $e->getMessage();
        $this->finishCheck($retObj, $userData);
        throw new Exception('transactionPurchase failed. '.$e->getMessage());
      }
      return $retObj;
    }

    /**
     * verify3D
     *
     * @param       $check3dId
     * @param       $config
     * @param array $userData
     *
     * @return bool|\stdClass
     * @throws \Exception
     */
    function verify3D($check3dId, $config, $userData = array())
    {
      if (!empty($config)) {
        $this->setConfig($config);
      }
      //echo __FUNCTION__.': '.$check3dId;
      try {
        $res = $this->verify3dNotification($check3dId);

        if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->verify3dNotification: '.print_r($res, TRUE)); // log result
        //mail('webmaster@web-dezign.de', __CLASS__.'->'.__FUNCTION__.'->$res', print_r($res,true));

        if ($res->sessionStatus == 'SUCCESS') {
          //$sessionId = $res->sessionId;
          //$res       = $this->transactionPurchase($sessionId);
          //if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->transactionPurchase: '.print_r($res, true)); // log result
          if ($res->transactionStatus == 'SUCCESS') {
            //$res->sessionId = $sessionId; // to have it in return value
            if (!empty($userData)) {
              $res = $this->finishCheck($res, $userData);
            }
            return $res;
          } else {
            $res->error = 1;
            return $res;
          }
        } else {
          throw new Exception('3D Secure failed: '.$res->transactionResultMessage);
        }

      } catch (Exception $e) {
        throw new Exception('verify3D failed. '.$e->getMessage());
      }

      return FALSE;
    }

    /**
     * refundTRX
     *
     * @param $order
     * @param $config
     *
     * @return bool|\stdClass
     * @throws \Exception
     */
    function refundTRX($order, $config)
    {
      if (empty($config['trxid'])) throw new Exception('TrxId cant be empty');
      if (empty($config['sessid'])) throw new Exception('SessId cant be empty');
      if (empty($config['amount'])) throw new Exception('Amount cant be empty');

      // switch to cent instead of eur
      $config['amount'] = (int)($config['amount'] * 100);

      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->config: '.print_r($config, TRUE));

      $this->setConfig($config);

      try {
        $sessionID     = $config['sessid'];
        $transactionId = $config['trxid'];
        $amount        = $config['amount'];
        $currency      = $config['currency'];
        $res           = $this->transactionRefund($sessionID, $transactionId, $amount, $currency);
        //mail('test@test.test', __CLASS__.'->'.__FUNCTION__.'->transactionRefund: ', print_r($res, true));
      } catch (Exception $e) {
        throw new Exception('transactionRefund failed. '.$e->getMessage());
      }

      $res->sessionId = $sessionID;
      return $res;
    }

    /**
     * bookSepa
     *
     * @param $userData
     * @param $config
     * @param $payData
     *
     * @return array|bool|\stdClass
     * @throws \Exception
     */
    function bookSepa($userData, $config, $payData)
    {
      $retObj = new stdClass();
      $this->setConfig($config);
      $this->switchPaymethod('sepa');

      $custID = $userData['id'];
      if (empty($custID)) $custID = 'guest-'.$userData['order_id']; // guest order
      $custID             = $this->unifyID($custID);
      $custObj            = new stdClass();
      $custObj->id        = $custID;
      $custObj->firstname = $userData['firstName'];
      $custObj->lastname  = $userData['lastName'];
      $custObj->email     = $userData['email'];
      $custObj->street    = $userData['street'];
      $custObj->zip       = $userData['zip'];
      $custObj->city      = $userData['city'];
      $custObj->country   = $userData['country'];
      $custObj->company   = $userData['company'];
      $custObj->lang      = strtolower($userData['lang']); //'de-DE';
      if (!empty($userData['ip'])) {
        $custObj->ip = $userData['ip'];
      } else {
        $custObj->ip = $_SERVER['REMOTE_ADDR'];
      }
      $orderID = $this->unifyID($userData['order_id']);
      if (empty($config['paytext']) && !empty($config['freetext'])) {
        $config['paytext'] = $config['freetext'];
      }
      $config['paytext'] = str_replace('{$orderid}', $orderID, $config['paytext']);
      $bookObj           = new stdClass();
      $bookObj->amount   = $userData['amount'];
      $bookObj->currency = $userData['currency'];
      $bookObj->orderID  = $orderID;
      $bookObj->title    = 'Order '.$orderID; // shown in control center
      $bookObj->paytext  = $config['paytext']; // shown on customer bank account
      $bookObj->payload  = array(
        'orderid'    => $orderID,
        'customerid' => $custID,
        'customerip' => $_SERVER['REMOTE_ADDR'],
      );
      if (!empty($config['payload'])) {
        foreach ($config['payload'] as $k => $v) {
          $bookObj->payload[$k] = $v;
        }
      }
      $custObj->payload    = $bookObj->payload;
      $bookObj->iban       = $payData['iban'];
      $bookObj->bic        = $payData['bic'];
      $bookObj->holder     = $payData['holder'];
      $bookObj->country    = $userData['country'];
      $bookObj->bankName   = $payData['bankName'];
      $bookObj->mandateRef = NULL; // SEPA Mandats Referenz, wird ansonsten erzeugt
      if (!empty($config['mandateRef'])) {
        $bookObj->mandateRef = $config['mandateRef'];
      }
      $bookObj->mandateSignDate = NULL; // Datum der Mandatserteilung, ansonsten sessionApprove-Datum
      $bookObj->mandateRecur    = 'ONEOFF'; // SEPA Mandat für einzelne - Standard für neue Mandate: "ONEOFF", oder wiederkehrende Zahlungen: "RECURRING" oder "FINAL"
      // default is ONEOFF - RECURRING only for abo payment
      if (!empty($config['isAbo'])) {
        $bookObj->mandateRecur = 'RECURRING';
      }

      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->custObj: '.print_r($custObj, TRUE));
      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->bookObj: '.print_r($bookObj, TRUE));
      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->config: '.print_r($config, TRUE));
      $retObj->custObj = $custObj;
      $retObj->bookObj = $bookObj;

      // create customer and session in one step but with update = 2 for customer update
      $customerCreated = FALSE;
      try {
        // customerGet check if customer exists
        $oldMandate = '';
        $res        = $this->customerGet($custObj->id);
        //mail('test@test.test', __CLASS__.'->'.__FUNCTION__ . '->customerGet('.$custObj->id.')', print_r($res, true));
        if (!empty($res->freeParams['mandateRef'])) {
          $oldMandate = $res->freeParams['mandateRef'];
        }
      } catch (Exception $e) {
        // if customer doesnt exists
        if ($e->getCode() == '3006') {
          try {
            $customerCreated  = TRUE;
            $payData['reuse'] = FALSE; // new customer cant reuse
            $this->customerCreate($custObj);
          } catch (Exception $e) {
            throw new Exception('customerCreate failed. '.$e->getMessage());
          }
          try {
            $this->bankaccountSet($custObj, $bookObj); // only mandatory if different data
          } catch (Exception $e) {
            throw new Exception('bankaccountSet failed. '.$e->getMessage());
          }
        } else {
          throw new Exception('customerGet failed. '.$e->getMessage());
        }
      }

      // check if there is something to set here! otherwise skip it.
      if (!empty($custObj->zip)) {
        try {
          $res = $this->addressSet($custObj->id, $custObj);
          //mail('test@test.test', __CLASS__.'->'.__FUNCTION__ . '->addressSet', print_r($res, true));
        } catch (Exception $e) {
          throw new Exception('addressSet failed. '.$e->getMessage());
        }
      }

      if ($customerCreated) { // if customer is new than set contact data
        try {
          $res = $this->contactDataSet($custObj->id, $custObj);
          //mail('test@test.test', __CLASS__.'->'.__FUNCTION__ . '->contactDataSet', print_r($res, true));
        } catch (Exception $e) {
          throw new Exception('contactDataSet failed. '.$e->getMessage());
        }
      }

      try {
        // get bank name
        if (empty($bookObj->bankName)) {
          $res = $this->bankaccountGet($custObj->id); // only mandatory if different data
          //mail('test@test.test', __CLASS__.'->'.__FUNCTION__ . '->bankaccountGet', print_r($res, true));
          $bookObj->bankName = $res->bankName;
          if (!empty($payData['reuse'])) {
            $bookObj->iban          = $res->iban;
            $bookObj->bic           = $res->bic;
            $bookObj->accountHolder = $res->accountHolder;
          }
        }
      } catch (Exception $e) {
//          if (!$customerCreated) {
//              $this->bankaccountSet($custObj, $bookObj); // only mandatory if different data
//              $this->bankaccountGet($custObj->id); // only mandatory if different data
//          }
      }

      // check change in bankdata and update bankaccount
      if ($res->iban != $bookObj->iban) {
        try {
          $this->bankaccountSet($custObj, $bookObj); // only mandatory if different data
        } catch (Exception $e) {
          throw new Exception('bankaccountSet failed. '.$e->getMessage(), $e->getCode());
        }
        try {
          $res = $this->bankaccountGet($custObj->id); // only mandatory if different data
          //mail('test@test.test', __CLASS__.'->'.__FUNCTION__ . '->bankaccountGet', print_r($res, true));
          $bookObj->bankName = $res->bankName;
        } catch (Exception $e) {
          throw new Exception('bankaccountGet failed. '.$e->getMessage(), $e->getCode());
        }
      }

      try {
        // check if old mandate exists and use it
        if (!empty($config['isAbo']) && !empty($oldMandate)) $bookObj->mandateRef = $oldMandate;
        // create new session
        $res = $this->sessionCreate($custObj, $bookObj);
        //mail('test@test.test', __CLASS__.'->'.__FUNCTION__.'->sessionCreate', print_r($res, true));
        $sessionId         = $res->sessionId;
        $retObj->resCreate = $res;
      } catch (Exception $e) {
        throw new Exception('sessionCreate failed. '.$e->getMessage(), $e->getCode());
      }
      try {
        if (!empty($config['isAbo']) && empty($oldMandate)) {
          $params = array('mandateRef' => $retObj->resCreate->mandateRef);
          $res    = $this->customerSet($custObj->id, $params);
          //mail('test@test.test', __CLASS__.'->'.__FUNCTION__ . '->customerSet', print_r($res, true));
        }
      } catch (Exception $e) {
        throw new Exception('customerSet failed. '.$e->getMessage());
      }

      $payRecurStates = array(
        'OOFF' => ' einmalig',
        'FRST' => ' erstmals',
        'RCUR' => ' wiederholt',
        'FNAL' => ' letztmalig',
      );
      $textDE         = 'Unsere Forderung in Höhe von {$amount} {$currency} ziehen wir mit der SEPA-Lastschrift zum Mandat Nr. {$mandateRef} '
        .'zu der Gläubiger-Identifikationsnummer {$creditorId} von Ihrem Konto IBAN {$iban} bei {$bankname} {$bic} '
        .'zum Fälligkeitstag {$payDateDE}{$payRecurSt} ein. Wir bitten Sie für Kontodeckung zu sorgen.';
      $textEN         = 'Please confirm the direct debit authorization. '
        .'We will deduct the payment amounting to {$amount} {$currency} using SEPA direct debit on the day it is due {$payDateEN}. '
        .'Please ensure that your account has sufficient funds.';
      if (!empty($config['textDE'])) $textDE = $config['textDE'];
      if (!empty($config['textEN'])) $textEN = $config['textEN'];
      if (empty($bookObj->bankName)) {
        $bookObj->bankName = 'Ihrer Bank';
      }

      $repl = array(
        '{$amount}'     => sprintf('%1.2f', ($retObj->resCreate->amount / 100)),
        '{$currency}'   => $retObj->resCreate->currency,
        '{$mandateRef}' => $retObj->resCreate->mandateRef,
        '{$creditorId}' => $retObj->resCreate->creditorId,
        '{$iban}'       => $bookObj->iban,
        '{$bankname}'   => $bookObj->bankName,
        '{$bic}'        => 'BIC '.$bookObj->bic,
        '{$payDateDE}'  => date('d.m.Y', strtotime($retObj->resCreate->payDate)),
        '{$payDateEN}'  => date('m/d/Y', strtotime($retObj->resCreate->payDate)),
        '{$payRecurSt}' => $payRecurStates[$retObj->resCreate->payRecurState],
      );
      if (empty($bookObj->bic)) $repl['{$bic}'] = ''; // if no bic so no bic text
      $retObj->mandateTextDE = strtr($textDE, $repl);
      $retObj->mandateTextEN = strtr($textEN, $repl);

      try {
        // approve payment
        $res = $this->sessionApprove($sessionId);
        //mail('test@test.test', __CLASS__.'->'.__FUNCTION__.'->sessionApprove', print_r($res, true));
        $retObj->resApprove = $res;
      } catch (Exception $e) {
        throw new Exception('sessionApprove failed. '.$e->getMessage());
      }
      if (empty($config['noFinish'])) {
        $retObj = $this->finishCheck($retObj, $userData);
      }

      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->retObj: '.print_r($retObj, TRUE));
      return $retObj;
    }

    /**
     * bookSepaShort
     *
     * @param $userData
     * @param $config
     * @param $payData
     *
     * @return mixed|stdClass
     * @throws Exception
     */
    function bookSepaShort($userData, $config, $payData)
    {
      $retObj = new stdClass();
      $res    = new stdClass();
      $this->setConfig($config);
      $this->switchPaymethod('sepa');

      //echo '<pre>'.print_r($payData,true).'</pre>';
      if (!empty($payData['remember'])) {
        $this->bakePayCookie($payData, $userData['id']);
      }

      $custID = $userData['id'];
      if (empty($custID)) $custID = 'guest-'.$userData['order_id']; // guest order
      $custID             = $this->unifyID($custID);
      $custObj            = new stdClass();
      $custObj->id        = $custID;
      $custObj->firstname = $userData['firstName'];
      $custObj->lastname  = $userData['lastName'];
      $custObj->email     = $userData['email'];
      $custObj->street    = $userData['street'];
      $custObj->zip       = $userData['zip'];
      $custObj->city      = $userData['city'];
      $custObj->country   = $userData['country'];
      $custObj->company   = $userData['company'];
      $custObj->lang      = strtolower($userData['lang']); //'de-DE';
      if (!empty($userData['ip'])) {
        $custObj->ip = $userData['ip'];
      } else {
        $custObj->ip = $_SERVER['REMOTE_ADDR'];
      }
      // fallback to firstname lastname if holder is empty
      if (empty($payData['holder'])) {
        $payData['holder'] = $userData['firstName'].' '.$userData['lastName'];
      }

      $orderID = $this->unifyID($userData['order_id']);
      if (empty($config['paytext']) && !empty($config['freetext'])) {
        $config['paytext'] = $config['freetext'];
      }
      $config['paytext'] = str_replace('{$orderid}', $orderID, $config['paytext']);
      $bookObj           = new stdClass();
      $bookObj->amount   = $userData['amount'];
      $bookObj->currency = $userData['currency'];
      $bookObj->orderID  = $orderID;
      $bookObj->title    = 'Order '.$orderID; // shown in control center
      $bookObj->paytext  = $config['paytext']; // shown on customer bank account
      $bookObj->payload  = array(
        'orderid'    => $orderID,
        'customerid' => $custID,
        'customerip' => $_SERVER['REMOTE_ADDR'],
      );
      if (!empty($config['payload'])) {
        foreach ($config['payload'] as $k => $v) {
          $bookObj->payload[$k] = $v;
        }
      }
      $custObj->payload    = $bookObj->payload;
      $bookObj->iban       = $payData['iban'];
      $bookObj->bic        = $payData['bic'];
      $bookObj->holder     = $payData['holder'];
      $bookObj->country    = $userData['country'];
      $bookObj->bankName   = $payData['bankName'];
      $bookObj->mandateRef = NULL; // SEPA Mandats Referenz, wird ansonsten erzeugt
      if (!empty($config['mandateRef'])) {
        $bookObj->mandateRef = $config['mandateRef'];
      }
      $bookObj->mandateSignDate = NULL; // Datum der Mandatserteilung, ansonsten sessionApprove-Datum
      $bookObj->mandateRecur    = 'ONEOFF'; // SEPA Mandat für einzelne - Standard für neue Mandate: "ONEOFF", oder wiederkehrende Zahlungen: "RECURRING" oder "FINAL"
      // default is ONEOFF - RECURRING only for abo payment
      if ($config['isAbo']) {
        $bookObj->mandateRecur = 'RECURRING';
      }
      if (!empty($config['payDate'])) {
        $bookObj->payDate = date('Y-m-d', strtotime('NOW + '.(int)$config['payDate'].' DAYS '));
      }
      if (isset($config['autogenSession'])) {
        $bookObj->autogenSession = (bool)$config['autogenSession'];
      }

      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->custObj: '.print_r($custObj, TRUE));
      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->bookObj: '.print_r($bookObj, TRUE));
      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->config: '.print_r($config, TRUE));
      $retObj->custObj = $custObj;
      $retObj->bookObj = $bookObj;

      try {
        // get bank data
        $res = $this->bankaccountGet($custObj->id); // only mandatory if different data
        //mail('test@test.test', __CLASS__.'->'.__FUNCTION__ . '->bankaccountGet', print_r($res, true));
        $bookObj->bankName = $res->bankName;
        if (!empty($payData['reuse'])) {
          $bookObj->iban          = $res->iban;
          $bookObj->bic           = $res->bic;
          $bookObj->accountHolder = $res->accountHolder;
        }
      } catch (Exception $e) {
        // retry backaccount without prefix
        if ($this->customPrefix != 'NONE') {
          $custID      = $userData['id'];
          $custObj->id = $custID;
          try {
            // get bank data
            $res = $this->bankaccountGet($custObj->id); // only mandatory if different data
            //mail('test@test.test', __CLASS__.'->'.__FUNCTION__ . '->bankaccountGet', print_r($res, true));
            $bookObj->bankName = $res->bankName;
            if (!empty($payData['reuse'])) {
              $bookObj->iban          = $res->iban;
              $bookObj->bic           = $res->bic;
              $bookObj->accountHolder = $res->accountHolder;
            }
          } catch (Exception $e) {
            throw new Exception('bankaccountGet failed. '.$e->getMessage());
          }
        }

        $payData['reuse'] = FALSE; // new customer cant reuse
        try {
          $this->customerCreate($custObj);
        } catch (Exception $e) {
          if ($e->getMessage() != 'customerId already exists' || $e->getCode() != '3006') {
            throw new Exception('customerCreate failed. '.$e->getMessage());
          }
        }
        try {
          $this->bankaccountSet($custObj, $bookObj); // only mandatory if different data
        } catch (Exception $e) {
          throw new Exception('bankaccountSet failed. '.$e->getMessage());
        }
        try {
          $this->addressSet($custObj->id, $custObj);
        } catch (Exception $e) {
          throw new Exception('addressSet failed. '.$e->getMessage());
        }
        try {
          $this->contactDataSet($custObj->id, $custObj);
        } catch (Exception $e) {
          throw new Exception('contactDataSet failed. '.$e->getMessage());
        }
      }

      // if iban not set
      if (!isset($res->iban)) $res->iban = '';

      // check change in bankdata and update bankaccount
      if ($res->iban != $bookObj->iban) {
        try {
          $this->bankaccountSet($custObj, $bookObj); // only mandatory if different data
        } catch (Exception $e) {
          throw new Exception('bankaccountSet failed. '.$e->getMessage(), $e->getCode());
        }
        try {
          $res = $this->bankaccountGet($custObj->id); // only mandatory if different data
          //mail('test@test.test', __CLASS__.'->'.__FUNCTION__ . '->bankaccountGet', print_r($res, true));
          $bookObj->bankName = $res->bankName;
        } catch (Exception $e) {
          throw new Exception('bankaccountGet failed. '.$e->getMessage(), $e->getCode());
        }
      }

      try {
        // create new session
        $res = $this->sessionCreate($custObj, $bookObj);
        //mail('test@test.test', __CLASS__.'->'.__FUNCTION__.'->sessionCreate', print_r($res, true));
        $sessionId         = $res->sessionId;
        $retObj->resCreate = $res;
      } catch (Exception $e) {
        throw new Exception('sessionCreate failed. '.$e->getMessage(), $e->getCode());
      }

      $payRecurStates = array(
        'OOFF' => ' einmalig',
        'FRST' => ' erstmals',
        'RCUR' => ' wiederholt',
        'FNAL' => ' letztmalig',
      );
      $textDE         = 'Unsere Forderung in Höhe von {$amount} {$currency} ziehen wir mit der SEPA-Lastschrift zum Mandat Nr. {$mandateRef} '
        .'zu der Gläubiger-Identifikationsnummer {$creditorId} von Ihrem Konto IBAN {$iban} bei {$bankname} {$bic} '
        .'zum Fälligkeitstag {$payDateDE}{$payRecurSt} ein. Wir bitten Sie für Kontodeckung zu sorgen.';
      $textEN         = 'Please confirm the direct debit authorization. '
        .'We will deduct the payment amounting to {$amount} {$currency} using SEPA direct debit on the day it is due {$payDateEN}. '
        .'Please ensure that your account has sufficient funds.';
      if (!empty($config['textDE'])) $textDE = $config['textDE'];
      if (!empty($config['textEN'])) $textEN = $config['textEN'];
      if (!empty($config['mandat_text_de'])) $textDE = $config['mandat_text_de'];
      if (!empty($config['mandat_text_en'])) $textEN = $config['mandat_text_en'];

      if (empty($bookObj->bankName)) {
        $bookObj->bankName = 'Ihrer Bank';
      }

      $repl = array(
        '{$amount}'     => sprintf('%1.2f', ($retObj->resCreate->amount / 100)),
        '{$currency}'   => $retObj->resCreate->currency,
        '{$mandateRef}' => $retObj->resCreate->mandateRef,
        '{$creditorId}' => $retObj->resCreate->creditorId,
        '{$iban}'       => $bookObj->iban,
        '{$bankname}'   => $bookObj->bankName,
        '{$bic}'        => 'BIC '.$bookObj->bic,
        '{$payDateDE}'  => date('d.m.Y', strtotime($retObj->resCreate->payDate)),
        '{$payDateEN}'  => date('m/d/Y', strtotime($retObj->resCreate->payDate)),
        '{$payRecurSt}' => $payRecurStates[$retObj->resCreate->payRecurState],
      );
      if (empty($bookObj->bic)) $repl['{$bic}'] = ''; // if no bic so no bic text
      $retObj->mandateTextDE = strtr($textDE, $repl);
      $retObj->mandateTextEN = strtr($textEN, $repl);

      try {
        // approve payment
        $res = $this->sessionApprove($sessionId);
        //mail('test@test.test', __CLASS__.'->'.__FUNCTION__.'->sessionApprove', print_r($res, true));
        $retObj->resApprove = $res;
      } catch (Exception $e) {
        throw new Exception('sessionApprove failed. '.$e->getMessage());
      }
      if (empty($config['noFinish'])) {
        $retObj = $this->finishCheck($retObj, $userData);
      }

      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->retObj: '.print_r($retObj, TRUE));
      return $retObj;
    }

    /**
     * refundSepa
     *
     * @param $order
     * @param $config
     *
     * @return bool|\stdClass
     * @throws \Exception
     */
    function refundSepa($order, $config)
    {
      if (empty($config['sessid'])) throw new Exception('SessId cant be empty');
      if (empty($config['amount'])) throw new Exception('Amount cant be empty');

      // switch to cent instead of eur
      $config['amount'] = (int)($config['amount'] * 100);

      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->config: '.print_r($config, TRUE));

      $this->setConfig($config);
      $this->switchPaymethod('sepa');

      try {
        $sessionID = $config['sessid'];
        $amount    = $config['amount'];
        $payText   = NULL;
        if (!empty($config['payText'])) {
          $payText = $config['payText'];
        }
        $res = $this->sessionRefund($sessionID, $amount, $payText);
        //mail('test@test.test', __CLASS__.'->'.__FUNCTION__.'->sessionRefund: ', print_r($res, true));
      } catch (Exception $e) {
        throw new Exception('sessionRefund failed. '.$e->getMessage());
      }

      $res->sessionId = $sessionID;
      return $res;
    }

    /**
     * bookPrepay
     *
     * @param $userData
     * @param $config
     *
     * @return \stdClass
     * @throws \Exception
     */
    function bookPrepay($userData, $config)
    {
      $retObj = new stdClass();
      $this->setConfig($config);
      $this->switchPaymethod('prepay');

      $custID = $userData['id'];
      if (empty($custID)) $custID = 'guest-'.$userData['order_id']; // guest order
      $custID             = $this->unifyID($custID);
      $custObj            = new stdClass();
      $custObj->id        = $custID;
      $custObj->firstname = $userData['firstName'];
      $custObj->lastname  = $userData['lastName'];
      $custObj->email     = $userData['email'];
      $custObj->street    = $userData['street'];
      $custObj->zip       = $userData['zip'];
      $custObj->city      = $userData['city'];
      $custObj->country   = $userData['country'];
      $custObj->company   = $userData['company'];
      $custObj->lang      = strtolower($userData['lang']); //'de-DE';
      if (!empty($userData['ip'])) {
        $custObj->ip = $userData['ip'];
      } else {
        $custObj->ip = $_SERVER['REMOTE_ADDR'];
      }

      $orderID = $this->unifyID($userData['order_id']);
      if (empty($config['paytext']) && !empty($config['freetext'])) {
        $config['paytext'] = $config['freetext'];
      }
      $config['paytext']   = str_replace('{$orderid}', $orderID, $config['paytext']);
      $bookObj             = new stdClass();
      $bookObj->amount     = $userData['amount'];
      $bookObj->currency   = $userData['currency'];
      $bookObj->orderID    = $orderID;
      $bookObj->title      = $config['paytext']; // shown in control center
      $bookObj->paytext    = $config['paytext']; // shown on customer bank account
      $bookObj->expireDays = $config['expiredays'];
      $bookObj->payload    = array(
        'orderid'    => $orderID,
        'customerid' => $custID,
        'customerip' => $_SERVER['REMOTE_ADDR'],
      );
      if (!empty($config['payload'])) {
        foreach ($config['payload'] as $k => $v) {
          $bookObj->payload[$k] = $v;
        }
      }
      $custObj->payload = $bookObj->payload;
      $bookObj->country = $userData['country'];
      if (!empty($config['sessionID'])) {
        $bookObj->sessionID = $config['sessionID'];
      }

      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->custObj: '.print_r($custObj, TRUE));
      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->bookObj: '.print_r($bookObj, TRUE));
      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->config: '.print_r($config, TRUE));
      $retObj->custObj = $custObj;
      $retObj->bookObj = $bookObj;

      try {
        // customerGet check if customer exists
        $res = $this->customerGet($custObj->id);
        //mail('test@test.test', __CLASS__.'->'.__FUNCTION__ . '->customerGet('.$custObj->id.')', print_r($res, true));
      } catch (Exception $e) {
        // if customer doesnt exists
        if ($e->getCode() == '3006') {
          try {
            $res = $this->customerCreate($custObj);
            //mail('test@test.test', __CLASS__.'->'.__FUNCTION__ . '->customerCreate', print_r($res, true));
          } catch (Exception $e) {
            throw new Exception('customerCreate failed. '.$e->getMessage());
          }
        } else {
          throw new Exception('customerGet failed. '.$e->getMessage());
        }
      }

      try {
        $res = $this->addressSet($custObj->id, $custObj);
        //mail('test@test.test', __CLASS__.'->'.__FUNCTION__ . '->addressSet', print_r($res, true));
      } catch (Exception $e) {
        throw new Exception('addressSet failed. '.$e->getMessage());
      }

      try {
        $res = $this->contactDataSet($custObj->id, $custObj);
        //mail('test@test.test', __CLASS__.'->'.__FUNCTION__ . '->contactDataSet', print_r($res, true));
      } catch (Exception $e) {
        throw new Exception('contactDataSet failed. '.$e->getMessage());
      }

      try {
        $res = $this->sessionCreate($custObj, $bookObj);
        //mail('test@test.test', __CLASS__.'->'.__FUNCTION__.'->sessionCreate', print_r($res, true));
        $retObj->resCreate = $res;
      } catch (Exception $e) {
        throw new Exception('sessionCreate failed. '.$e->getMessage());
      }

      $textDE = 'Sehr geehrte Kundin, sehr geehrter Kunde,<br><br>'
        .'vielen Dank für Ihre Bestellung vom {$orderDateDE}. micropayment übernimmt für uns die Zahlungsabwicklung.<br><br>'
        .'Bitte überweisen Sie {$amount} {$currency} bis zum {$dueDateDE} auf die folgende Kontoverbindung, um den Kauf abzuschließen.<br><br>'
        .'Kontoinhaber: {$accountHolder}<br>'
        .'IBAN: {$iban}<br>'
        .'BIC: {$bic}<br>'
        .'Bank: {$bankName}<br>'
        .'Gesamtbetrag: {$amount} {$currency}<br>'
        .'Verwendungszweck: {$payText}';
      $textEN = 'Dear Customer,<br><br>'
        .'Thank you for purchase on {$orderDateEN}. Payment processing for us is provided by micropayment.<br><br>'
        .'Please make transfer in the amount of {$amount} {$currency} by {$dueDateEN} in order to complete your purchase. '
        .'Payment should be made to the following bank account.<br><br>'
        .'Beneficiary: {$accountHolder}<br>'
        .'IBAN: {$iban}<br>'
        .'BIC: {$bic}<br>'
        .'Bank: {$bankName}<br>'
        .'Total amount: {$amount} {$currency}<br>'
        .'Payment purpose: {$payText}';
      if (!empty($config['textDE'])) $textDE = $config['textDE'];
      if (!empty($config['textEN'])) $textEN = $config['textEN'];
      if (!empty($config['prepay_text_de'])) $textDE = $config['prepay_text_de'];
      if (!empty($config['prepay_text_en'])) $textEN = $config['prepay_text_en'];

      $repl = array(
        '{$amount}'        => sprintf('%1.2f', ($retObj->resCreate->amount / 100)),
        '{$currency}'      => $retObj->resCreate->currency,
        '{$dueDateDE}'     => date('d.m.Y', strtotime($retObj->resCreate->dueDate)),
        '{$dueDateEN}'     => date('m/d/Y', strtotime($retObj->resCreate->dueDate)),
        '{$orderDateDE}'   => date('d.m.Y'),
        '{$orderDateEN}'   => date('m/d/Y'),
        '{$iban}'          => $retObj->resCreate->iban,
        '{$bic}'           => $retObj->resCreate->bic,
        '{$accountHolder}' => $retObj->resCreate->accountHolder,
        '{$bankName}'      => $retObj->resCreate->bankName,
        '{$payText}'       => $retObj->resCreate->payText,
      );
      if (empty($retObj->resCreate->bic)) $repl['{$bic}'] = ''; // if no bic so no bic text
      $retObj->prepayTextDE = strtr($textDE, $repl);
      $retObj->prepayTextEN = strtr($textEN, $repl);
      $retObj->rawData      = $repl; // return raw data
      $retObj               = $this->finishCheck($retObj, $userData);

      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->retObj: '.print_r($retObj, TRUE));
      return $retObj;
    }

    /**
     * refundPrepay
     *
     * @param $order
     * @param $config
     *
     * @return bool|\stdClass
     * @throws \Exception
     */
    function refundPrepay($order, $config)
    {
      if (empty($config['sessid'])) throw new Exception('SessId cant be empty');
      if (empty($config['amount'])) throw new Exception('Amount cant be empty');

      // switch to cent instead of eur
      $config['amount'] = (int)($config['amount'] * 100);

      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->config: '.print_r($config, TRUE));

      $this->setConfig($config);
      $this->switchPaymethod('prepay');

      try {
        $sessionID       = $config['sessid'];
        $refundAmount    = $config['amount'];
        $bookObj         = new stdClass();
        $bookObj->amount = ($order->get_total() * 100) - $refundAmount; // in CENT
        $res             = $this->sessionChange($sessionID, $bookObj);
        mail('test@test.test', __CLASS__.'->'.__FUNCTION__.'->sessionChange: ', print_r($res, TRUE));
      } catch (Exception $e) {
        throw new Exception('sessionChange failed. '.$e->getMessage());
      }

      $res->sessionId = $sessionID;
      return $res;
    }

    /**
     * getPaymentWindowParams
     *
     * @param $order
     * @param $config
     * @param $pm
     *
     * @return string
     */
    function getPaymentWindowParams($userData, $config, $pm)
    {
      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->config('.$pm.'): '.print_r($config, TRUE));
      if (!array_key_exists($pm, $this->mcpayEventURLs)) {
        throw new Exception('Unknown paymethod '.$pm);
      }

      if (empty($config['paytext']) && !empty($config['freetext'])) {
        $config['paytext'] = $config['freetext'];
      }

      $params = array(
        'currency'    => $userData['currency'],
        'project'     => $config['project'],
        'testmode'    => (int)json_decode($config['testMode']),
        'theme'       => $config['theme'],
        'gfx'         => $config['gfx'],
        'bgcolor'     => $config['bgcolor'],
        'bggfx'       => $config['bggfx'],
        'amount'      => $userData['amount'],
        //'title'    => $_SERVER['HTTP_HOST'].' - ORDER '.$userData['order_id'],
        //'paytext'  => str_replace('{$orderid}', $userData['order_id'], $config['paytext']),
        'title'       => $userData['order_id'],
        'paytext'     => str_replace('{$orderid}', $userData['order_id'], $config['paytext']),
        //'currency' => $userData['currency'],
        'orderid'     => $userData['order_id'],
        'producttype' => 'cart',

        'mp_user_email'     => $userData['email'],
        'mp_user_firstname' => $userData['firstName'],
        'mp_user_surname'   => $userData['lastName'],
        'mp_user_address'   => $userData['street'],
        'mp_user_zip'       => $userData['zip'],
        'mp_user_city'      => $userData['city'],
        'mp_user_country'   => $userData['country'],
        'mp_user_id'        => $userData['id'],
        'lang'              => strtolower($userData['lang']),
      );
      if (in_array($pm, array('sofort', 'obt'))) {
        $params['lang'] = strtoupper($params['lang']);
      }
      if (!empty($config['producttype'])) {
        $params['producttype'] = $config['producttype'];
      }
      if (in_array($params['mp_user_email'], $this->testAccounts)) {
        $params['testmode']            = 1; // test account book in testmode on
        $config['extraParams']['note'] = 'Test account used !!';
        $params['title']               = '!TEST! '.$params['title'];
        $params['paytext']             = '!TEST! '.$params['paytext'];
      }
      if (!empty($userData['trxid'])) {
        $config['extraParams']['trxid'] = $userData['trxid'];
      }
      if (!empty($userData['oid'])) {
        $config['extraParams']['oid'] = $userData['oid'];
      }
      if (!empty($userData['respurl'])) {
        $config['extraParams']['respurl'] = $userData['respurl'];
      }

      // check if extra params given
      if (!empty($config['extraParams'])) {
        foreach ($config['extraParams'] as $k => $v) {
          // check if already filled. no overwrite !
          if (empty($params[$k])) {
            $params[$k] = $v;
          }
        }
      }
      if ($userData['doFinish']) {
        $this->finishCheck($pm, $userData);
      }
      $query  = http_build_query($params, '', '&');
      $seal   = md5($query.$config['accessKey']);
      $url    = str_replace('micropayment.de', $config['suffix'], $this->mcpayEventURLs[$pm]);
      $return = $url.'?'.$query.'&seal='.$seal;
      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->$params: '.print_r($params, TRUE).$return);
      return $return;
    }

    /**
     * getBaseJS
     *
     * @param int  $tokenize
     * @param int  $autobill
     * @param bool $formSubmit
     * @param bool $debug
     * @param null $formID
     *
     * @return string
     */
    public function getBaseJS($tokenize = 1, $autobill = 0, $formSubmit = TRUE, $debug = FALSE, $formID = NULL)
    {
      $jscript = '<script src="'.$this->mcpayBridgeURL.'"
			project 		= "'.$this->project.'"
			testmode 		= "'.(int)json_decode($this->testMode).'"
			panformat 	= "1"
			tokenize 		= "'.$tokenize.'"
			error 			= "'.$this->error_color.'"
			error_off 	= "'.$this->error_off_color.'"
			prefix 			= "'.$this->prefixCCard.'"
			autobill 		= "'.$autobill.'"
    ';
      if (!empty($formID)) {
        $jscript .= ' form = "'.$formID.'" ';
      }
      if ($formSubmit) {
        $jscript .= ' form_submit = "auto" ';
      } else {
        $jscript .= ' form_submit = "off" ';
      }
      if ($debug) $jscript .= ' debug = 1 ';
      //$jscript.= ' trxid = "1234567890-'.$tokenize.'"';
      $jscript .= '></script>';
      return $jscript;
    }

    /**
     * getBaseJSInline
     *
     * @param int  $tokenize
     * @param int  $autobill
     * @param bool $formSubmit
     * @param bool $debug
     * @param null $formID
     *
     * @return string
     */
    public function getBaseJSInline($tokenize = 1, $autobill = 0, $formSubmit = FALSE, $debug = FALSE, $formID = NULL)
    {
      $jscript = 'var s = document.createElement("script");
      s.setAttribute( "src", "'.$this->mcpayBridgeURL.'" );
      s.setAttribute( "project", "'.$this->project.'" );
      s.setAttribute( "testmode", "'.(int)json_decode($this->testMode).'" );
      s.setAttribute( "panformat", "1" );
      s.setAttribute( "tokenize", "'.$tokenize.'" );
      s.setAttribute( "error", "'.$this->error_color.'" );
      s.setAttribute( "error_off", "'.$this->error_off_color.'" );
      s.setAttribute( "prefix", "'.$this->prefixCCard.'" );
      s.setAttribute( "autobill", "'.$autobill.'" );
    ';
      if (!empty($formID)) {
        $jscript .= ' s.setAttribute( "form", "'.$formID.'" ); ';
      }
      if ($formSubmit) {
        $jscript .= ' s.setAttribute( "form_submit", "auto" ); ';
      } else {
        $jscript .= ' s.setAttribute( "form_submit", "off" ); ';
      }
      if ($debug) $jscript .= ' s.setAttribute( "debug", "1" ); ';
      $jscript .= ' document.head.appendChild(s); ';
      return $jscript;
    }

    /**
     * getInlineJS
     * get outer part of the inline script
     * to inject the jscript via jscript
     *
     * @param string $mcpayJSInline - return of getBaseJSIline()
     *
     * @return string
     */
    private function getInlineJS($mcpayJSInline)
    {
      $jscript = '//console.log("getInlineJS");
      // remember first event
      var CHDone = false;
      // get '.$this->prefixCCard.'holder input to set onfocus event
      var CH = document.getElementById("'.$this->prefixCCard.'holder");
      // add onfocus event
      CH.onfocus = function(){
        // console.log("mcpay CH started");
        // check first event done
        if (CHDone) return;
        doMIPA();
      };
      var CP = document.getElementById("'.$this->prefixCCard.'pan");
      // add onclick event
      CP.onclick = function(){
        // console.log("mcpay CP started");
        // check first event done
        if (CHDone) return;
        doMIPA();
      };
      
      function doMIPA()
      {
        // set first event done
        CHDone = true;
        '.$mcpayJSInline.'
        // when micropayment script loaded
        s.onload = function(){
          // add styling script for iframe fields
          var s2 = document.createElement("script");
          s2.setAttribute( "src", "'.$this->plugin_urljs.'mcpay_formstyle_ccard_iframe'.$this->fileExtJS.'" );
          document.head.appendChild(s2);
        };
      }
    ';
      return $jscript;
    }

    /**
     * getStartJS
     *
     * @param bool $inline
     * @param bool $formSubmit
     * @param bool $debug
     * @param null $formID
     *
     * @return string
     */
    public function getStartJS($inline = FALSE, $formSubmit = FALSE, $debug = FALSE, $formID = NULL)
    {
      if ($inline) {
        $mcpayJSInline = $this->getBaseJSInline(1, 0, $formSubmit, $debug, $formID);
        $jscript       = $this->getInlineJS($mcpayJSInline);
      } else {
        $jscript = $this->getBaseJS(1, 0, $formSubmit, $debug, $formID).
          '<script src="'.$this->plugin_urljs.'mcpay_formstyle_ccard_check'.$this->fileExtJS.'"></script>'.
          '<script src="'.$this->plugin_urljs.'mcpay_formstyle_ccard_iframe'.$this->fileExtJS.'"></script>'.
          '<link href="'.$this->plugin_urlcss.'mcpay_formstyle_ccard'.$this->fileExtCSS.'" rel="stylesheet" />';
      }
      return $jscript;
    }

    /**
     * getFinishJS
     *
     * @param        $sessionID
     * @param        $responseURL
     * @param string $method
     * @param bool   $formSubmit
     * @param bool   $debug
     * @param null   $formID
     *
     * @return string
     */
    public function getFinishJS($sessionID, $responseURL, $method = 'Purchase', $formSubmit = FALSE, $debug = FALSE, $formID = NULL)
    {
      $possibleVals = array('Purchase', 'Authorization');
      if (!in_array($method, $possibleVals)) $method = 'Purchase';
      //mail('te@st.de', __CLASS__.'->'.__FUNCTION__, 'getFinishJS');
      $jscript = $this->getBaseJS(0, 0, $formSubmit, $debug, $formID).
        '<script type="text/javascript">
				var prefix = "'.$this->prefixCCard.'";
				getElem = function(name)
        {
          return document.getElementById(name);
        }
        getElem(prefix+"error").onclick = function()
        {
          history.back();
        }
        getElem(prefix+"modal").onclick = function()
        {
          history.back();
        }
				onError = function(e)
        {
          console.log(e.msg);
          getElem(prefix+"error").innerHTML = e.msg+"";
          getElem(prefix+"error").setAttribute("class", "show");
        }
        window.onload = function(){
				  Micropayment.addEvent("error", onError);
				  Micropayment.session("'.$sessionID.'", "'.$responseURL.'", "", "dieId", "'.$method.'");
				  Micropayment.process();
				}
			</script>';
      $output  = '<div id="'.$this->prefixCCard.'error" class="hide"></div>';
      $output  .= '<div id="'.$this->prefixCCard.'load" class="show"></div>';
      $output  .= '<div id="'.$this->prefixCCard.'modal" class="show"></div>';
      $output  .= $jscript;
      $output  .= '<link href="'.$this->plugin_urlcss.'mcpay_formstyle_ccard" rel="stylesheet" />';
      return $output;
    }

    /**
     * getCCardTPL
     *
     * @param       $filename
     * @param array $options
     * @param array $prefillData
     * @param array $translation
     *
     * @return bool|string
     */
    public function getCCardTPL($filename, $options = array(), $prefillData = array(), $translation = array())
    {
      if (empty($filename)) return FALSE;

      $formTPL = file_get_contents($filename);

      //dd($options);
      if (empty($prefillData) && !empty($options['showRemember'])) {
        $UID = NULL;
        if (!empty($options['UID'])) {
          $UID = $options['UID'];
        }
        $prefillData = $this->checkPayCookieCC($UID);
        //dd($prefillData);
      }

      $imgStyle = 'float: right; padding-right: 9px; margin-top: -25px;';
      if (!empty($options['imgStyle'])) {
        $imgStyle = $options['imgStyle'];
      }
      if (!empty($options['hideImg'])) {
        $imgStyle = 'display: none;';
      }
      $imgURL = 'https://www.micropayment.de/resources/?what=img&group=cc&show=type-n.1';
      if (!empty($options['imgURL'])) {
        $imgURL = $options['imgURL'];
      }

      $years = '<option value="0">Year</option>';
      for ($i = date('y'); $i <= (date('y') + $options['futureYears']); $i++) {
        $selected = '';
        if ($options['preSeletActDate'] && date('y') == $i) {
          $selected = ' selected';
        }
        if (!empty($options['prefillPost'][$this->prefixCCard.'year']) && $options['prefillPost'][$this->prefixCCard.'year'] == $i) {
          $selected = ' selected';
        }
        $years .= '<option value="'.$i.'" '.$selected.'>'.($i + 2000).'</option>';
      }

      $month = '<option value="0">Month</option>';
      for ($i = 1; $i <= 12; $i++) {
        $selected = '';
        if ($options['preSeletActDate'] && date('m') == $i) {
          $selected = ' selected';
        }
        if (!empty($options['prefillPost'][$this->prefixCCard.'month']) && $options['prefillPost'][$this->prefixCCard.'month'] == $i) {
          $selected = ' selected';
        }
        $month .= '<option value="'.$i.'"'.$selected.'>';
        if (!empty($options['showMonthNames'])) {
          $month .= '#'.date('M', mktime(0, 0, 0, $i)).'#'.' ('.sprintf('%02d', $i).')';
        } else {
          //$month .= date('m', mktime(0, 0, 0, $i));
          $month .= sprintf('%02d', $i); // use numbers only because date function makes problems at the end of month
        }
        $month .= '</option>';
      }

      $showNewForm = 'show';
      $checked     = 'checked'; // preselect first saved
      if (!empty($options['preselectNewCard'])) {
        $showNewForm = 'hide';
        $checked     = '';
      }
      $checkedOP = 'checked';
      if ($checked == 'checked') $checkedOP = '';
      //$remember = '';
      //if ($options['showRemember']) {
      $remember = '<br><div id="'.$this->prefixCCard.'remember" ';
      if (!empty($options['showRemember'])) {
        $remember .= 'class="show"';
      } else {
        $remember .= 'class="hide"';
      }
      $checkRemember = '';
      if (!empty($options['checkRemember'])) {
        $checkRemember = 'checked';
      }
      $remember .= '><input type="checkbox" id="'.$this->prefixCCard.'remember_chk" name="'.$this->prefixCCard.'remember" value="1" '.$checkRemember.'>';
      $remember .= '<label for="'.$this->prefixCCard.'remember_chk">Remember Data</label></div>';
      //}
      $savedCards = '';
      $newCard    = '';
      if (!empty($prefillData)) {
        $newCard .= '<div class="reuse">';
        $newCard .= '<input type="radio" id="'.$this->prefixCCard.'reuse" name="'.$this->prefixCCard.'reuse" value="new" onclick="handleCCardRadioClick(this);" '.$checkedOP.'>';
        $newCard .= '<label for="'.$this->prefixCCard.'reuse">Use new Card</label><br>';
        $newCard .= '</div>';
        //$checked = 'checked';
        $savedCards = $this->getCCardCards($prefillData, $options);

        if ($showNewForm == 'show')
          $showNewForm = 'hide';
        else
          $showNewForm = 'show';
      }

      $repl = array(
        'PREFIX_holder_val'                      => '',
        '<option>Year</option>'                  => $years,
        '<option>Month</option>'                 => $month,
        '<div id="PREFIX_remember"></div>'       => $remember,
        '<div id="PREFIX_reuse"></div>'          => $savedCards,
        '<div id="PREFIX_newcard"></div>'        => '<div id="'.$this->prefixCCard.'newcard">'.$newCard.'</div>',
        '<div id="PREFIX_newform" class="show">' => '<div id="'.$this->prefixCCard.'newform" class="'.$showNewForm.'">',
        'PREFIX_'                                => $this->prefixCCard,
        '{IMG_STYLE}'                            => $imgStyle,
        '{IMG_URL}'                              => $imgURL,
      );
      // mark fields
      if (!empty($options['classes'])) {
        foreach ($options['classes'] as $k => $v) {
          $repl['PREFIX_'.$k.'_class'] = $v;
        }
      }
      // prefill fields
      if (!empty($options['prefillPost'])) {
        foreach ($options['prefillPost'] as $k => $v) {
          $k               = str_replace($this->prefixCCard, 'PREFIX_', $k);
          $repl[$k.'_val'] = $v;
        }
      }
      $formTPL = strtr($formTPL, $repl);
      if (!empty($translation)) {
        $formTPL = strtr($formTPL, $translation);
      }

      return $formTPL;
    }

    /**
     * getCCardCards
     *
     * @param       $prefillData
     * @param array $options
     *
     * @return string
     */
    public function getCCardCards($prefillData, $options = array())
    {
      $crlf = "\n";
      $tab  = '  ';

      $showNewForm = 'show';
      $checked     = 'checked'; // preselect first saved
      if (!empty($options['preselectNewCard'])) {
        $showNewForm = 'hide';
        $checked     = '';
      }
      $savedCards = '';
      foreach ($prefillData as $k => $v) {
        //$checked    = $v['checked'];
        if (empty($v['pan'])) continue;

        if (empty($v['token'])) $v['token'] = 'old';
        if (empty($v['pan'])) $v['pan'] = '';
        if (empty($v['brand'])) $v['brand'] = '';
        if (empty($v['month'])) $v['month'] = date('m');
        if (empty($v['year'])) $v['year'] = date('Y');

        $savedCards .= '<div class="reuse prefill">';
        if (!empty($options['cardSelect'])) {
          $savedCards .= '<span class="'.$this->prefixCCard.'reuseselect">';
          $savedCards .= '<input type="radio" name="'.$this->prefixCCard.'reuse" value="'.$v['token'].'" onclick="handleCCardRadioClick(this);" '.$checked.'>';
        } else {
          $savedCards .= '<span class="'.$this->prefixCCard.'reuse_select">';
          if (empty($options['showCardOnly'])) {
            $savedCards .= '<input type="radio" id="'.$this->prefixCCard.'reuse_old" name="'.$this->prefixCCard.'reuse" value="'.$v['token'].'" onclick="handleCCardRadioClick(this);" '.$checked.'>';
            $savedCards .= '<label for="'.$this->prefixCCard.'reuse_old">Use old Card</label><br>';
          }
        }
        $savedCards .= '<div id="'.$this->prefixCCard.'reuse_card" class="'.$showNewForm.'">';

        if (!empty($options['cardStyle'])) {
          if (!empty($options['specialBrand'])) {
            $brand = substr($v['holder'], 0, strpos($v['holder'], '(') - 1);
          } else {
            $brand = $v['brand'];
          }
          if (empty($brand)) {
            $cardType = 'visa';
            if (substr($v['pan'], 0, 1) == '5') $cardType = 'mastercard';
            $brand = $cardType;
          }
          if (!empty($options['specialHolder'])) {
            $holder = substr($v['holder'], strpos($v['holder'], '(') + 1, -1);
          } else {
            $holder = $v['holder'];
          }
          $savedCards .= '<div class="'.$this->prefixCCard.'reusecard">';
          //$savedCards .= '<img src="' . $options['imgPath'] . $brand . '.png">';
          $savedCards .= '<img src="'.$this->logoURLs[$brand].'">';

          $savedCards .= '<span class="'.$this->prefixCCard.'cardlabel">NAME:</span>'.$holder.'<br>';
          $savedCards .= '<span class="'.$this->prefixCCard.'cardlabel">PAN:</span>'.$v['pan'].'<br>';
          $savedCards .= '<span class="'.$this->prefixCCard.'cardlabel">EXPIRE:</span>'.$v['month'].'/'.$v['year'].'<br>';
          $savedCards .= '</div>';
        } else {
          $savedCards .= 'Holder: '.$v['holder'].' CardNo: '.$v['pan'].' Expire: '.$v['month'].'/'.$v['year'];
        }
        $savedCards .= '</span>';
        if (!empty($options['showRemove'])) {
          $savedCards .= '<span style="display: inline-block; float: right; margin-top: -5px">';
          $savedCards .= '<input type="checkbox" name="'.$this->prefixCCard.'remove[]" value="'.$k.'"> Remove';
          $savedCards .= '</span>';
        }
        $savedCards .= '</div>';
        $savedCards .= '</div>';
        //$checked = '';
      }
      return $savedCards;
    }

    /**
     * getSepaTPL
     *
     * @param       $filename
     * @param array $options
     * @param array $prefillData
     * @param array $translation
     *
     * @return bool|false|string
     */
    public function getSepaTPL($filename, $options = array(), $prefillData = array(), $translation = array())
    {
      if (empty($filename)) return FALSE;
      $crlf = "\n";
      $tab  = '  ';

      $formTPL = file_get_contents($filename);

      if (empty($prefillData) && !empty($options['showRemember'])) {
        $UID = NULL;
        if (!empty($options['UID'])) {
          $UID = $options['UID'];
        }
        $prefillData = $this->checkPayCookie($UID);
        //dd($prefillData);
      }

      $imgStyle = 'float: right; padding-right: 9px; margin-top: -25px;';
      if (!empty($options['imgStyle'])) {
        $imgStyle = $options['imgStyle'];
      }
      if (!empty($options['hideImg'])) {
        $imgStyle = 'display: none;';
      }
      $imgURL = 'https://www.micropayment.de/resources/?what=img&group=dbt&show=type-n.1';
      if (!empty($options['imgURL'])) {
        $imgURL = $options['imgURL'];
      }

      $showNewForm = 'show';
      $checked     = 'checked'; // preselect first saved
      if (!empty($options['preselectNewCard'])) {
        $showNewForm = 'hide';
        $checked     = '';
      }
      $checkedOP = 'checked';
      if ($checked == 'checked') $checkedOP = '';
      //$remember = '';
      //if ($options['showRemember']) {
      $remember = '<br><div id="'.$this->prefixSepa.'remember" ';
      if (!empty($options['showRemember'])) {
        $remember .= 'class="show"';
      } else {
        $remember .= 'class="hide"';
      }
      $checkRemember = '';
      if (!empty($options['checkRemember'])) {
        $checkRemember = 'checked';
      }
      $remember .= '><input type="checkbox" id="'.$this->prefixSepa.'remember_chk" name="'.$this->prefixSepa.'remember" value="1" '.$checkRemember.'>';
      $remember .= '<label for="'.$this->prefixSepa.'remember_chk">Remember Data</label></div>';
      //}
      $savedCards = '';
      $newCard    = '';
      if (!empty($prefillData)) {
        $newCard    .= '<div class="reuse">'.$crlf;
        $newCard    .= $tab.$tab.'<input type="radio" id="'.$this->prefixSepa.'reuse" name="'.$this->prefixSepa.'reuse" value="new" onclick="handleSepaRadioClick(this);" '.$checkedOP.'>'.$crlf;
        $newCard    .= $tab.$tab.'<label for="'.$this->prefixSepa.'reuse">Use new Card</label><br>'.$crlf;
        $newCard    .= $tab.'</div>'.$crlf;
        $checked    = 'checked';
        $savedCards = $this->getSepaCards($prefillData, $options);

        if ($showNewForm == 'show')
          $showNewForm = 'hide';
        else
          $showNewForm = 'show';
      }

      $repl = array(
        'PREFIX_holder_val'                      => '',
        'PREFIX_iban_val'                        => '',
        'PREFIX_bic_val'                         => '',
        'PREFIX_holder_class'                    => '',
        'PREFIX_iban_class'                      => '',
        'PREFIX_bic_class'                       => '',
        '<div id="PREFIX_remember"></div>'       => $remember,
        '<div id="PREFIX_reuse"></div>'          => $savedCards,
        '<div id="PREFIX_newcard"></div>'        => '<div id="'.$this->prefixSepa.'newcard">'.$newCard.'</div>'.$crlf,
        '<div id="PREFIX_newform" class="show">' => '<div id="'.$this->prefixSepa.'newform" class="'.$showNewForm.'">',
        'PREFIX_'                                => $this->prefixSepa,
        '{IMG_STYLE}'                            => $imgStyle,
        '{IMG_URL}'                              => $imgURL,
      );
      if ($this->debug && FALSE) {
        if (strpos($_SERVER['REMOTE_ADDR'], '127.0.0') !== FALSE) {
          $repl['PREFIX_holder_val'] = 'Donald Duck';
          $repl['PREFIX_iban_val']   = 'DE89888888881234567890';
          $repl['PREFIX_bic_val']    = 'TESTDE00XXX';
        }
      }
      // mark fields
      if (!empty($options['classes'])) {
        foreach ($options['classes'] as $k => $v) {
          $repl['PREFIX_'.$k.'_class'] = $v;
        }
      }
      // prefill fields
      if (!empty($options['prefillPost'])) {
        foreach ($options['prefillPost'] as $k => $v) {
          $k               = str_replace($this->prefixSepa, 'PREFIX_', $k);
          $repl[$k.'_val'] = $v;
        }
      }
      $formTPL = strtr($formTPL, $repl);
      if (!empty($translation)) {
        $formTPL = strtr($formTPL, $translation);
      }

      return $formTPL;
    }

    /**
     * getSepaCards
     *
     * @param       $prefillData
     * @param array $options
     *
     * @return string
     */
    public function getSepaCards($prefillData, $options = array())
    {
      $crlf = "\n";
      $tab  = '  ';

      $showNewForm = 'show';
      $checked     = 'checked'; // preselect first saved
      if (!empty($options['preselectNewCard'])) {
        $showNewForm = 'hide';
        $checked     = '';
      }
      $savedCards = '';
      foreach ($prefillData as $k => $v) {
        $savedCards .= '<div class="reuse prefill">'.$crlf;
        $savedCards .= $tab.$tab.'<span class="'.$this->prefixSepa.'reuse_select">'.$crlf;
        if (empty($options['showCardOnly'])) {
          $savedCards .= $tab.$tab.$tab.'<input type="radio" id="'.$this->prefixSepa.'reuse_old" name="'.$this->prefixSepa.'reuse" value="'.$k.'" onclick="handleSepaRadioClick(this);" '.$checked.'>'.$crlf;
          $savedCards .= $tab.$tab.$tab.'<label for="'.$this->prefixSepa.'reuse_old">Use old Card</label><br>'.$crlf;
        }
        $savedCards .= $tab.$tab.$tab.'<div id="'.$this->prefixSepa.'reuse_card" class="'.$showNewForm.'">'.$crlf;
        if (!empty($options['cardStyle'])) {
          $brand      = 'ec';
          $savedCards .= $tab.$tab.$tab.$tab.'<div class="'.$this->prefixSepa.'reusecard">'.$crlf;
          //$savedCards .= '<img src="' . $options['imgPath'] . $brand . '.png">';
          $savedCards .= $tab.$tab.$tab.$tab.$tab.'<img src="'.$this->logoURLs[$brand].'">'.$crlf;
          $savedCards .= $tab.$tab.$tab.$tab.$tab.'<span class="'.$this->prefixSepa.'cardlabel">NAME:</span>'.$v['holder'].'<br>'.$crlf;
          $savedCards .= $tab.$tab.$tab.$tab.$tab.'<input type="hidden" name="'.$this->prefixSepa.'reuse_holder" value="'.$v['holder'].'">'.$crlf;
          $savedCards .= $tab.$tab.$tab.$tab.$tab.'<span class="'.$this->prefixSepa.'cardlabel">IBAN:</span>'.$v['iban'].'<br>'.$crlf;
          $savedCards .= $tab.$tab.$tab.$tab.$tab.'<input type="hidden" name="'.$this->prefixSepa.'reuse_iban" value="'.$v['iban'].'">'.$crlf;
          $savedCards .= $tab.$tab.$tab.$tab.$tab.'<span class="'.$this->prefixSepa.'cardlabel">BIC:</span>'.$v['bic'].'<br>'.$crlf;
          $savedCards .= $tab.$tab.$tab.$tab.$tab.'<input type="hidden" name="'.$this->prefixSepa.'reuse_bic" value="'.$v['bic'].'">'.$crlf;
          $savedCards .= $tab.$tab.$tab.$tab.'</div>'.$crlf;
        } else {
          $savedCards .= 'Holder: '.$v['holder'].' IBAN: '.$v['iban'].' BIC: '.$v['bic'].$crlf;
        }
        if (!empty($options['showRemove'])) {
          $savedCards .= $tab.$tab.$tab.'<span style="display: block;  margin-top: -30px;margin-left: 215px;text-align: left;margin-bottom: 10px;">'.$crlf;
          $savedCards .= $tab.$tab.$tab.'<input type="checkbox" name="'.$this->prefixSepa.'remove[]" value="'.$k.'"> Remove'.$crlf;
          $savedCards .= $tab.$tab.$tab.'</span>'.$crlf;
        }
        $savedCards .= $tab.$tab.$tab.'</div>'.$crlf;
        $savedCards .= $tab.$tab.'</span>'.$crlf;
        $savedCards .= $tab.'</div>'.$crlf;
      }

      return $savedCards;
    }

    /**
     * bakePayCookie
     *
     * @param $payData
     * @param $UID
     *
     * @return bool
     */
    private function bakePayCookie($payData, $UID)
    {
      //echo '<pre>'.print_r($payData,true).'</pre>';
      if (!json_decode($payData['remember'])) {
        unset($_COOKIE['mcpay_paycookie'][$UID]);
        if ($payData['reuse'] != '1') {
          setcookie('mcpay_paycookie['.$UID.']', NULL, -1, '/');
        }
        return TRUE;
      } // if not remember, dont set cookie
      if (!empty($payData) && !empty($payData['iban']) && !empty($UID)) {
        $data = json_encode($payData);
        setcookie('mcpay_paycookie['.$UID.']', $data, time() + 60 * 60 * 24 * 30, '/');
      }
      return TRUE;
    }

    /**
     * checkPayCookie
     *
     * @param $UID
     *
     * @return array
     */
    private function checkPayCookie($UID)
    {
      //echo '<pre>'.print_r($_COOKIE,true).'</pre>';
      if (empty($UID)) return array();
      if (empty($_COOKIE['mcpay_paycookie'])) return array();

      if (!empty($_COOKIE['mcpay_paycookie'][$UID])) {
        $cookieData = json_decode($_COOKIE['mcpay_paycookie'][$UID], TRUE);
        //echo '<pre>'.print_r($cookieData,true).'</pre>';

        // check real saved data at mcp
        $this->switchPaymethod('sepa');
        $res = $this->bankaccountGet($UID);
        //echo '<pre>'.print_r($res,true).'</pre>';
        $cookieData['iban']     = $res->iban;
        $cookieData['bic']      = $res->bic;
        $cookieData['country']  = $res->country;
        $cookieData['bankName'] = $res->bankName;
        $cookieData['holder']   = $res->accountHolder;

        if (!empty($cookieData['iban'])) { // mask IBAN
          $newIBAN = substr($cookieData['iban'], 0, 4);
          for ($i = 0; $i <= strlen($cookieData['iban']) - 8; $i++) {
            $newIBAN .= '*';
          }
          $newIBAN            .= substr($cookieData['iban'], -4);
          $cookieData['iban'] = $newIBAN;
        }
        //echo '<pre>'.$UID.print_r($cookieData,true).'</pre>';
        return array(1 => $cookieData);
      }
      return array();
    }

    /**
     * bakePayCookieCC
     *
     * @param $payData
     * @param $UID
     *
     * @return bool
     */
    private function bakePayCookieCC($payData, $UID)
    {
      if (!json_decode($payData['remember'])) {
        unset($_COOKIE['mcpay_paycookie_cc'][$UID]);
        if ($payData['token'] != 'reuse') {
          setcookie('mcpay_paycookie_cc['.$UID.']', NULL, -1, '/');
        }
        return TRUE;
      } // if not remember, dont set cookie
      if (!empty($payData) && !empty($payData['pan']) && !empty($UID)) {
        $data = json_encode($payData);
        setcookie('mcpay_paycookie_cc['.$UID.']', $data, time() + 60 * 60 * 24 * 30, '/');
      }
      return TRUE;
    }

    /**
     * checkPayCookieCC
     *
     * @param $UID
     *
     * @return array
     */
    private function checkPayCookieCC($UID = NULL)
    {
      if (empty($UID)) return array();
      if (empty($_COOKIE['mcpay_paycookie_cc'])) return array();

      if (!empty($_COOKIE['mcpay_paycookie_cc'][$UID])) {
        $cookieData = json_decode($_COOKIE['mcpay_paycookie_cc'][$UID], TRUE);
        // check real saved data at mcp
        $this->switchPaymethod('ccard');
        $res = $this->creditcardDataGet($UID);
        //echo '<pre>'.print_r($res,true).'</pre>';
        $cookieData['pan']   = $res->number;
        $cookieData['month'] = $res->expiryMonth;
        $cookieData['year']  = $res->expiryYear;
        $cookieData['type']  = $res->type;
        //echo '<pre>'.$UID.print_r($cookieData,true).'</pre>';
        return array(1 => $cookieData);
      }
      return array();
    }

    /**
     * isLocal
     *
     * @return bool
     */
    public static function isLocal()
    {
      if (file_exists(__DIR__.'/../islocal.txt')) return TRUE;
      return FALSE;
    }

    /**
     * getLogoURL
     *
     * @return string
     */
    public static function getLogoURL()
    {
      if (self::isLocal()) {
        return self::$mcpayLogoURLDEV;
      }
      return self::$mcpayLogoURL;
    }

    /**
     * getLogoPicURL
     *
     * @return string
     */
    public static function getLogoPicURL()
    {
      if (self::isLocal()) {
        return self::$mcpayLogoPicURLDEV;
      }
      return self::$mcpayLogoPicURL;
    }

    /**
     * getVersionURL
     *
     * @return string
     */
    public static function getVersionURL()
    {
      //return self::$mcpayVersionCheckURL;
      if (self::isLocal()) {
        return self::$mcpayVersionCheckURLDEV;
      }
      return self::$mcpayVersionCheckURL;
    }

    /**
     * getCSSFile
     *
     * @param      $sn
     * @param bool $withHeader
     *
     * @return false|string
     */
    public function getCSSFile($sn, $withHeader = TRUE)
    {
      if (empty($sn)) return FALSE;
      if ($withHeader) header('Content-Type: text/css');
      $file = __DIR__.'/../css/'.$sn.'.css';
      //echo $file;
      return file_get_contents($file);
    }

    /**
     * getJSFile
     *
     * @param      $sn
     * @param bool $withHeader
     *
     * @return false|string
     */
    public function getJSFile($sn, $withHeader = TRUE)
    {
      if (empty($sn)) return FALSE;
      if ($withHeader) header('Content-Type: application/javascript');
      $file = __DIR__.'/../js/'.$sn.'.js';
      //echo $file;
      return file_get_contents($file);
    }

    /**
     * markFinished
     *
     * @param     $refid
     * @param int $status
     *
     * @throws Exception
     */
    public function markFinished($refid, $status = 10)
    {
      $dat = array(
        'action' => 'update',
        'refid'  => $refid,
      );
      $url = self::getLogoPicURL().'?d='.base64_encode(serialize($dat)).'&s='.$status;
      $this->fireCall($url, array());
    }

    /**
     * finishCheck
     *
     * @param     $retObj
     * @param     $userData
     *
     * @return mixed
     */
    private function finishCheck($retObj, $userData)
    {
      $status = '1';
      if (is_object($retObj)
        && !empty($retObj->trxInfo)
        && is_object($retObj->trxInfo)
        && !empty($retObj->trxInfo->error)
        && $retObj->trxInfo->sessionStatus == 'SUCCESS'
        && $retObj->trxInfo->transactionStatus == 'SUCCESS') {
        $userData['chksum']['pm'] = 'ccard';
        if ($retObj->trxInfo->secureStatus != 'START3D') {
          // Success without 3D Secure
          $status = '10';
        } else {
          // 3D Secure started
          $status = '2';
        }
      } else if (is_object($retObj)
        && !empty($retObj->resCreate)
        && is_object($retObj->resCreate)
        && !empty($retObj->resCreate->error)
        && $retObj->resCreate->status == 'INIT') {
        $userData['chksum']['comment'] = $retObj->resCreate->payToken;
        $userData['chksum']['pm']      = 'prepay';
        if (!empty($retObj->resCreate->mandateRef)) {
          $userData['chksum']['pm'] = 'sepa';
        }
        $status = '3'; // prepay started
        if ($retObj->resApprove->status == 'APPROVED') {
          $status = '10'; // sepa paid
        }
      } else if (is_object($retObj)
        && !empty($retObj->sessionStatus)
        && $retObj->sessionStatus == 'SUCCESS'
        && !empty($retObj->transactionStatus)
        && $retObj->transactionStatus == 'SUCCESS') {
        // Success with 3D Secure
        $status                        = '11';
        $userData['chksum']['comment'] = '3DSecure';
      } else {
        // Error happened
        $status = '0';
        if (!empty($userData['doFinish'])) {
          $userData['chksum']['pm']         = $retObj;
          $userData['chksum']['plugintype'] = 'WEB';
          $status                           = '1';
        } else {
          if (!empty($retObj->trxInfo)) {
            if (empty($userData['chksum']['pm'])) $userData['chksum']['pm'] = 'ccard';
            $userData['chksum']['comment'] = $retObj->trxInfo->transactionResultMessage.' ('.$retObj->trxInfo->transactionResultCode.')';
          } else if (!empty($retObj->errorMessage)) {
            $userData['chksum']['comment'] = $retObj->errorMessage;
          } else if (!empty($retObj->transactionResultMessage)) {
            $userData['chksum']['comment'] = $retObj->transactionResultMessage.' ('.$retObj->transactionResultCode.')';
          } else {
            $userData['chksum']['comment'] = print_r($retObj, TRUE);
          }
        }
      }
      //$userData['chksum']['refid'] = $userData['order_id'];
      $url = self::getLogoPicURL().'?d='.base64_encode(serialize($userData['chksum'])).'&s='.$status;
      $this->fireCall($url, array());
      return $retObj;
    }

    /**
     * findLogFile
     *
     * @param $logPath
     * @param $logFile
     *
     * @return bool|string
     */
    function findLogFile($logPath, $logFile = NULL)
    {
      $files = array();
      if ($handle = opendir($logPath)) {
        while (FALSE !== ($entry = readdir($handle))) {
          if (!empty($logFile)) {
            if (substr($entry, 0, strlen($logFile)) == $logFile) {
              return $logPath.$entry;
            }
          } else {
            if (strtolower(substr($entry, -4)) == '.log') $files[] = $entry;
          }
        }
        closedir($handle);
      }
      sort($files);
      if (!empty($files)) return $files;
      return FALSE;
    }

    /**
     * readLogFile
     *
     * @param      $logFile
     * @param bool $returnRAW
     *
     * @return array|bool|false|string
     */
    function readLogFile($logFile, $returnRAW = FALSE, $date = NULL)
    {
      if (empty($logFile)) return FALSE;

      $repl    = array(
        ': Array'           => ': XArray',
        ': stdClass Object' => ': XstdClass Object',
        'script'            => 'pre',
        '<div'              => '&lt;div',
        'div>'              => 'div&gt;',
      );
      $content = file_get_contents($logFile);
      if ($returnRAW) return nl2br($content);
      $content = strtr($content, $repl);
      //$content = htmlentities($content);
      //echo $content;

      $delimiter = $date.'T';
      //$p = explode("\n\n", $content);
      $p = explode($delimiter, $content);
      //echo '<pre>'.print_r($p,true).'</pre>'; exit();
      $content = array();
      foreach ($p as $k => $v) {
        //echo substr($v, 19, 12)."\n";
        if (substr($v, 19 - strlen($delimiter), 12) == '+00:00 DEBUG') {
          $pp                       = explode('+00:00 DEBUG ', $v);
          $ppp                      = explode(": X", $pp[1]);
          $content[$pp[0]][$ppp[0]] = $ppp[1].'<br>'.$ppp[2];
          //$content[] = $ppp[0].'<hr>'.$ppp[1].'<br>'.$ppp[2];
        } else {
          continue;
        }
      }
      //echo '<pre>'.print_r($content,true).'</pre>'; exit();
      $output = '';
      foreach ($content as $time => $dat) {
        $output .= '<h3>'.$time.'</h3>';
        foreach ($dat as $k => $v) {
          $output .= '<pre>'.htmlentities($k).'<br>'.str_replace('&lt;br&gt;', '<br>', htmlentities($v)).'</pre>';
        }
        $output .= '<hr>';
      }

      $styleGray   = '<b style="background-color: lightgrey; padding:2px; color: black">';
      $styleGreen  = '<b style="background-color: green; padding:2px; color: white">';
      $styleYellow = '<b style="background-color: yellow; padding:2px; color: black">';
      $styleOrange = '<b style="background-color: orange; padding:2px; color: white">';
      $styleRed    = '<b style="background-color: red; padding:2px; color: white">';
      $stylePurple = '<b style="background-color: mediumpurple; padding:2px; color: white">';
      $styleBlue   = '<b style="background-color: blue; padding:2px; color: white">';
      $styleBlack  = '<b style="background-color: black; padding:2px; color: white">';
      $styleCyan   = '<b style="background-color: cyan; padding:2px; color: black">';
      $styleEnd    = '</b>';
      $repl        = array(
        'BOOKING'              => $styleGreen.'BOOKING'.$styleEnd,
        'Mickey Maus'          => $styleGreen.'Mickey Maus'.$styleEnd,
        'REVERSED'             => $styleOrange.'REVERSED'.$styleEnd,
        'STOP'                 => $styleRed.'STOP'.$styleEnd,
        'REVERSAL'             => $styleOrange.'REVERSAL'.$styleEnd,
        'BACKPAY'              => $styleBlue.'BACKPAY'.$styleEnd,
        'RECHARGED'            => $styleBlue.'RECHARGED'.$styleEnd,
        'responseAction'       => $styleBlack.'responseAction'.$styleEnd,
        'error:'               => $styleRed.'error:'.$styleEnd,
        '[status] =&gt; ERROR' => $styleRed.'[status] =&gt; ERROR'.$styleEnd,
        'trxInfo'              => $styleCyan.'trxInfo'.$styleEnd,
      );
      $output      = strtr($output, $repl);

      $statOutput = '';
      foreach ($repl as $k => $v) {
        $count = preg_match_all('#'.preg_quote($k, '#').'#', $output);
        if ($count > 0) {
          $statOutput .= '<div style="display: inline-block; margin-right: 5px;">'.str_replace($k, $count.'x '.$k, $repl[$k]).'</div>';
        }
      }
      echo $statOutput;
      return $output;
    }

    /**
     * readerLogFile
     *
     * @param      $logFile
     * @param int  $readBlock
     * @param bool $returnRAW
     *
     * @return bool|string
     */
    function readerLogFile($logFile, $readBlock = 0, $returnRAW = FALSE)
    {
      if (empty($logFile)) return FALSE;
      $maxlen = filesize($logFile);
      $offset = 0;
      if (!empty($readBlock)) {
        $offset = $maxlen - $readBlock;
      }

      $linedelimiter = '{CRLF}';
      $timedelimiter = '{DEL}';
      $content       = file_get_contents($logFile, FALSE, NULL, $offset);
      if ($returnRAW) return nl2br($content);
      $content = preg_replace('/[^ ](\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})/', $linedelimiter.'$1'.$timedelimiter, $content);
      //echo $content; exit();

      $lines = explode($linedelimiter, $content);
      //echo '<pre>'.print_r($lines,true).'</pre>'; exit();
      $content = array();
      foreach ($lines as $k => $v) {
        $p    = explode($timedelimiter, $v);
        $time = str_replace('[', '', $p[0]);
        if (empty($p[1])) continue;
        $cont             = $p[1];
        $content[$time][] = $cont;
      }
      //echo '<pre>'.print_r($content,true).'</pre>'; exit();
      krsort($content);
      $output = '';
      foreach ($content as $time => $dat) {
        $output .= '<h3>'.$time.'</h3>';
        foreach ($dat as $k => $v) {
          $output .= '<pre>'.str_replace('&lt;br&gt;', '<br>', htmlentities($v)).'</pre>';
        }
        $output .= '<hr>';
      }

      $filter = array(
        'sql:'                   => 'Purple',
        'bookPrepay'             => 'Yellow',
        'bookCCard'              => 'Yellow',
        'bookSepa'               => 'Yellow',
        'forwardURL'             => 'Blue',
        'errorMessage'           => 'Red',
        'getPaymentWindowParams' => 'Cyan',
        'PAID'                   => 'Green',
        'OVERPAID'               => 'Orange',
        'CHARGED'                => 'Orange',
        'REVERSED'               => 'Red',
        'RECHARGED'              => 'Orange',
        'FAILED'                 => 'Red',
        'EXPIRED'                => 'Red',
        'FATAL'                  => 'Red',
        'ERROR'                  => 'Red',
        'Notification'           => 'Gray',
        'notify'                 => 'Gray',
        'response'               => 'Gray',
      );

      $replWritings = array(
        'forwardUrl' => 'forwardURL',
        'start3d'    => 'START3D',
      );

      $styleBase = '<b style="background-color: #BGCOL#; color: #FGCOL#; padding:2px; border: 1px solid #BCOL#" onClick="window.find(\'#SEARCHTEXT#\')">#SEARCHTEXT#</b>';
      $styles    = array(
        'Gray'     => 'lightgrey;black',
        'Darkgray' => 'darkgrey;white',
        'Green'    => 'green;white',
        'Yellow'   => 'yellow;black;black',
        'Orange'   => 'orange;black;black',
        'Red'      => 'red;white',
        'Purple'   => 'mediumpurple;white',
        'Blue'     => 'blue;white',
        'Black'    => 'black;white',
        'Cyan'     => 'lightcyan;black;black',
      );
      $repl      = array();
      foreach ($filter as $t => $c) {
        $p = explode(';', $styles[$c]);
        if (empty($p[1])) $p[1] = 'black'; // default font
        if (empty($p[2])) $p[2] = 'white'; // default border
        $tmp      = array(
          '#BGCOL#'      => $p[0],
          '#FGCOL#'      => $p[1],
          '#BCOL#'       => $p[2],
          '#SEARCHTEXT#' => $t,
        );
        $repl[$t] = strtr($styleBase, $tmp);
      }
      $output = strtr($output, $replWritings);
      $output = strtr($output, $repl);

      $statOutput = '';
      foreach ($repl as $k => $v) {
        $count = preg_match_all('#'.preg_quote($k, '#').'#', $output);
        if ($count > 0) {
          $statOutput .= '<div style="display: inline-block; margin-right: 5px;" onClick="window.find(\''.$k.'\')">'.str_replace($k, ($count / 2).'x '.$k, $repl[$k]).'</div>';
        }
      }
      echo $statOutput;
      return $output;
    }

    /**
     * camelCase
     * convert a camel_case_string to camelCaseString
     *
     * @param $string
     * @param $sep
     *
     * @return string
     */
    public function camelCase($string, $sep = '_')
    {
      $p         = explode($sep, $string);
      $newString = '';
      foreach ($p as $k) {
        $newString .= ucfirst($k);
      }
      return lcfirst($newString);
    }

    public function getNotification($post, $data = array())
    {
      $return           = new \stdClass();
      $return->error    = FALSE;
      $return->errorMSG = '';
      $return->mpStatus = '';
      $func             = '';
      $crlf             = "\r\n";
      if (empty($post)) $post = $_POST;
      if (empty($post)) $post = $_GET;
      $amount           = 0;
      $currency         = 'EUR';
      $importantActions = array(
        'transactionCreate',
        'transactionRefund',
        'transactionChargeback',
        'transactionBackpay',
        'sessionStatus',
      );
      //echo '<pre>'.print_r($post,true).'</pre>';

      if (!empty($post['secretfield'])) {
        if ($this->secretfield != $post['secretfield']) {
          $return->mpStatus = 'status=error'.$crlf.'message=Communication is insecure.';
        }
      }

      // api notifications
      $isAPI = FALSE;
      if (!empty($_POST['action'])) {
        $isAPI = TRUE;
        if (!in_array($_POST['action'], $importantActions)) {
          $return->error    = TRUE;
          $return->errorMSG = 'Action not important!';
          $return->mpStatus = 'status=ok'.$crlf.'message=Action '.$_POST['action'];
          if (!empty($_POST['type'])) {
            $return->mpStatus .= '->'.$_POST['type'];
          }
          $return->mpStatus .= ' ignored.';
        } else {
          if ($_POST['action'] == 'sessionStatus') {
            $post['function'] = $_POST['status'];
            $post['auth']     = 'isPaid';
          } else {
            $post['function'] = $_POST['type'];
            $post['auth']     = $_POST['transactionId'];
          }
          $post['trxid']    = 'seeOrder';
          $post['testmode'] = $_POST['testMode'];
          $post['amount']   = $_POST['paidAmount'];
          $post['orderid']  = $_POST['sessionId'];
        }
        if (!empty($return->error)) {
          if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->API return'.print_r($return, TRUE));
          echo $return->mpStatus;
        }
      }

      // check mandatory params
      if (empty($post['function'])) {
        $return->error    = TRUE;
        $return->mpStatus = 'status=error'.$crlf.'message=Function missing.';
      } else {
        $func = $post['function'];
      }

      if ($func == 'getScript') {
        return $this->getScriptAction();
      }
      if ($func == 'getStyle') {
        return $this->getStyleAction();
      }
      if ($func == 'getImage') {
        return $this->getImageAction();
      }
      if ($func == 'handle3D') {
        $res = $this->handle3DAction($post);
        if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->handle3DAction'.print_r($res, TRUE));
        return $res;
      }
      if ($func == 'handleNO3D') {
        $res = $this->handleNO3DAction($post);
        if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->handleNO3DAction'.print_r($res, TRUE));
        return $res;
      }


      if ($func != 'custinfo') {
        if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->post'.print_r($post, TRUE));

        if (empty($post['orderid'])) {
          $return->error    = TRUE;
          $return->mpStatus = 'status=error'.$crlf.'message=OrderID missing.';
        } else {
          $orderId = $post['orderid'];
        }

        if (empty($post['auth'])) {
          $return->error    = TRUE;
          $return->mpStatus = 'status=error'.$crlf.'message=Auth missing.';
        } else {
          $authId = $post['auth'];
        }

        if (empty($post['amount'])) {
          $return->error    = TRUE;
          $return->mpStatus = 'status=error'.$crlf.'message=Amount missing.';
        } else {
          $amount = $post['amount'];
        }

        if (!empty($post['currency'])) {
          $currency = $post['currency'];
        }

        $theAmount = sprintf('%1.2f', $amount / 100).' '.$currency;

//                if (!isset($post['testmode'])){
//                    $return->error = TRUE;
//                    $return->mpStatus = 'status=error'.$crlf.'message=Testmode missing.';
//                } else {
//                    $testmode = $post['testmode'];
//                }

        if (empty($post['trxid'])) {
          $return->error    = TRUE;
          $return->mpStatus = 'status=error'.$crlf.'message=TRX ID missing.';
        } else {
          $trxid = $post['trxid'];
        }

        $oid = NULL;
        if (!empty($post['oid'])) $oid = $post['oid'];

        $respurl = '';
        if (empty($post['intresp']) && in_array($func, array('billing', 'init'))) {
          if (!empty($post['respurl'])) {
            $respurl = $post['respurl'];
          }
        }

      } else {
        if (empty($post['ak'])) {
          $return->error    = TRUE;
          $return->mpStatus = 'status=error'.$crlf.'message=Missing KEY!';
        } else {
          $ak = $post['ak'];
        }
      }

      if (!empty($return->error)) {
        if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->return'.print_r($return, TRUE));
        echo $return->mpStatus;
      }
      // default error after pre-checks
      $return->mpStatus = 'status=error'.$crlf.'message=Unknown Error';

      switch ($func) {

        // status call
        case 'custinfo':
          if ($this->accessKey == $ak) {
            if (!empty($post['log'])) {
              echo $this->readerLogFile($this->log->error_log);
              exit();
            } else {
              $infos = 'Shop Infos: '.$data['shopInfo'].'<br>';
              $infos .= 'Plugin Infos: '.$data['pluginInfo'].'<br>';
              $infos .= '<textarea>'.print_r(get_object_vars($this), TRUE).'</textarea>';

              echo '<pre>'.print_r($infos, TRUE).'</pre>';
              exit();
            }
          } else {
            $return->error    = TRUE;
            $return->mpStatus = 'status=error'.$crlf.'message=Access denied!';
          }
          break;

        case 'BOOKING':
          $return->state    = 'paid';
          $return->comment  = date('d.m.Y H:i:s').': '.$return->state.' Amount: '.$theAmount;
          $return->mpStatus = 'status=ok'.$crlf.'message='.$return->state.' done.'.$crlf;
          break;

        case 'CREATE':
        case 'CHANGE':
          $return->state    = 'reopen';
          $return->comment  = date('d.m.Y H:i:s').': '.$return->state.' Amount: '.$theAmount;
          $return->mpStatus = 'status=ok'.$crlf.'message='.$return->state.' done.'.$crlf;
          break;

        // billing call
        case 'billing':
        case 'init':
          if (empty($data['successURL'])) {
            $return->error    = TRUE;
            $return->errorMSG = 'successURL cant be empty!';
          } else {
            $return->state = 'paid';
            $return->paidAmount = $amount;
            $return->openAmount = 0;
            if (in_array($func, array('init'))) {
              $return->state = 'init';
              $return->paidAmount = 0; // not really paid yet
              $return->openAmount = $amount;
            }
            if ($isAPI) {
              if (!empty($data['sessionID']) && $data['sessionID'] != '__sessionId__') {
                try {
                  $trx = $this->sessionGet($data['sessionID']);
                  if ($trx->status == 'SUCCESS') {
                    $return->error = FALSE;
                  }
                } catch (Exception $e) {
                  $return->error    = TRUE;
                  $return->mpStatus = 'status=fail'.$crlf.'message='.$e->getMessage().$crlf;
                }
              }
            }

            if (empty($return->error)) {
              $return->comment  = date('d.m.Y H:i:s').': '.$return->state.' Amount: '.$theAmount;
              $return->mpStatus = 'status=ok'.$crlf // ok  | error
                .'url='.$data['successURL'].$crlf
                .'target=_top'.$crlf
                .'forward=1'.$crlf
                .'message=OK'.$crlf;
            }
          }
          break;

        case 'payin': // payin prepayment
        case 'PAYIN':
        case 'PAID':
          $return->state = 'paid';

          if ($isAPI) {
            if (!empty($data['sessionID']) && $data['sessionID'] != '__sessionId__') {
              try {
                $trx = $this->sessionGet($data['sessionID']);
                if ($trx->status == 'SUCCESS') {
                  $return->error = FALSE;
                  // @TODO: CHECK THIS ON API CALL
                  if ($trx->paystatus == 'OPEN') {
                    if ($trx->openamount > 0) {
                      $return->state      = 'part_paid';
                      $return->openAmount = $trx->openamount;
                      $return->paidAmount = $trx->paidamount + $trx->amount;
                    }
                  } else if (in_array($trx->paystatus, array('PAID', 'OVERPAID'))) {
                    $return->state      = 'part_paid';
                    $return->openAmount = 0;
                    $return->paidAmount = $trx->paidamount + $trx->amount;
                  }
                }
              } catch (Exception $e) {
                $return->error    = TRUE;
                $return->mpStatus = 'status=fail'.$crlf.'message='.$e->getMessage().$crlf;
              }
            }
          } else {
            if ($post['paystatus'] == 'OPEN') {
              if ($post['openamount'] > 0) {
                $return->state      = 'part_paid';
                $return->openAmount = $post['openamount'];
                $return->paidAmount = $post['paidamount'] + $post['amount'];
              }
            } else if (in_array($post['paystatus'], array('PAID', 'OVERPAID'))) {
              $return->state      = 'part_paid';
              $return->openAmount = 0;
              $return->paidAmount = $post['paidamount'] + $post['amount'];
            }
          }

          $return->comment  = date('d.m.Y H:i:s').': '.$return->state.' Amount: '.$theAmount;
          $return->mpStatus = 'status=ok'.$crlf.'message='.$return->state.' done.'.$crlf;
          break;

        case 'expire': // expire prepayment
          $return->state    = 'expire';
          $return->comment  = date('d.m.Y H:i:s').': '.$return->state.' Amount: '.$theAmount;
          $return->mpStatus = 'status=ok'.$crlf.'message='.$return->state.' done.'.$crlf;
          break;

        case 'storno': // chargeback sepa
        case 'transactionChargeback': // chargeback ccard
          $return->state    = 'cancel';
          $return->comment  = date('d.m.Y H:i:s').': '.$return->state.' Amount: '.$theAmount;
          $return->mpStatus = 'status=ok'.$crlf.'message='.$return->state.' done.'.$crlf;
          break;

        case 'backpay': // chargeback reversal
        case 'REFUNDREVERSAL'; // refund reversal
        case 'transactionBackpay': // backpay ccard
          $return->state    = 'backpay';
          $return->comment  = date('d.m.Y H:i:s').': '.$return->state.' Amount: '.$theAmount;
          $return->mpStatus = 'status=ok'.$crlf.'message='.$return->state.' done.'.$crlf;
          break;

        case 'refund': // refund sepa
        case 'REFUND': // refund prepayment
        case 'transactionRefund': // refund ccard
          if ($isAPI) {
            if (!empty($data['sessionID']) && $data['sessionID'] != '__sessionId__') {
              try {
                $trx         = $this->sessionGet($data['sessionID']);
                $orderAmount = $trx->amount;

                $return->state = 'refund';
                if ($amount < $orderAmount) {
                  $return->state = 'part_refund';
                }
                $return->comment  = date('d.m.Y H:i:s').': '.$return->state.' Amount: '.$theAmount;
                $return->mpStatus = 'status=ok'.$crlf.'message='.$return->state.' done.'.$crlf;
              } catch (Exception $e) {
                $return->error    = TRUE;
                $return->mpStatus = 'status=fail'.$crlf.'message='.$e->getMessage().$crlf;
              }
            }
          } else {
            $return->state = 'refund';
            $return->openAmount = $data['openAmount'];
            $return->refundAmount = $amount;
            if (!empty($data['openAmount']) && $data['openAmount'] > $amount){
              $return->state = 'part_refund';
            }
            $return->comment  = date('d.m.Y H:i:s').': '.$return->state.' Amount: '.$theAmount;
            $return->mpStatus = 'status=ok'.$crlf.'message='.$return->state.' done.'.$crlf;
          }
          break;

        case 'quit': // reversal SEPA only
          $return->state    = 'quit';
          $return->comment  = date('d.m.Y H:i:s').': '.$return->state.' Amount: '.$theAmount;
          $return->mpStatus = 'status=ok'.$crlf.'message='.$return->state.' done.'.$crlf;
          break;
      }

      $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->return: '.print_r($return, TRUE));

      return $return;
    }

    /**
     * getStyleAction
     * read and echo the style css with fitting header content type
     *
     * @return string|void
     */
    public function getStyleAction()
    {
      $req = array();
      if (!empty($_GET)) $req = $_GET;

      if (empty($req['sn'])) {
        if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->Req empty: '.print_r($req['sn'], 1));
        return 'param sn missing!';
      }

      $filename = MCPAY_BASE_PATH.'css/'.$req['sn'].'.css';
      if (!file_exists($filename)) {
        if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->File not found: '.print_r($filename, 1));
        return 'file not found!';
      }
      $content = file_get_contents($filename);

      header('Content-Type: text/css');
      echo $content;
      exit();
    }

    /**
     * getScriptAction
     * read and echo the script css with fitting header content type
     *
     * @param $payformid
     *
     * @return string|void
     */
    public function getScriptAction($payformid = 'confirm--form')
    {
      $req = array();
      if (!empty($_GET)) $req = $_GET;

      if (empty($req['sn'])) {
        if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->Req empty: '.print_r($req['sn'], 1));
        return 'param sn missing!';
      }

      $filename = MCPAY_BASE_PATH.'js/'.$req['sn'].'.js';
      if (!file_exists($filename)) {
        if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->File not found: '.print_r($filename, 1));
        return 'file not found!';
      }

      $content = file_get_contents($filename);
      if (strpos($filename, 'check.js') !== FALSE) {
        $repl    = array(
          'payformid'             => $payformid,
          'mcpay_card_token-form' => $payformid,
        );
        $content = strtr($content, $repl);
      }

      header('Content-Type: application/javascript');
      echo $content;
      exit();
    }

    /**
     * getImageAction
     * read and echo the image css with fitting header content type
     *
     * @return string|void
     */
    public function getImageAction()
    {
      $req = array();
      if (!empty($_GET)) $req = $_GET;

      if (empty($req['sn'])) {
        if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->Req empty: '.print_r($req['sn'], 1));
        return 'param sn missing!';
      }

      $filename = MCPAY_BASE_PATH.'img/'.$req['sn'].'.png';
      if (!file_exists($filename)) {
        if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->File not found: '.print_r($filename, 1));
        return 'file not found!';
      }
      $content = file_get_contents($filename);

      header('Content-Type: image/png');
      echo $content;
      exit();
    }

    /**
     * handle3DAction
     *
     * @param $post
     *
     * @return bool|\stdClass
     */
    public function handle3DAction($post)
    {
      $returnURL = urldecode($post['returnURL']);
      $errorURL  = urldecode($post['errorURL']);

      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->post'.print_r($_REQUEST, TRUE));

      if (!empty($_REQUEST['check3dId'])) {
        $check3dId = $_REQUEST['check3dId'];
      } else {
        $res           = new \stdClass();
        $res->error    = 1;
        $res->errorMSG = '3D Secure failed!';
      }

      if (!$res->error) {
        try {
          $res = $this->verify3D($check3dId);
        } catch (\Exception $e) {
          $res           = new \stdClass();
          $res->error    = 1;
          $res->errorMSG = $e->getMessage();
        }

        if (isset($res->error) && $res->error == 0) {
          $res->state        = 'paid';
          $res->redirect     = 1;
          $res->redirect_url = $returnURL;
        } else {
          $res->state        = 'fail';
          $res->redirect     = 1;
          $res->redirect_url = $errorURL;
        }
      }
      if ($this->debug) $this->log->debug(__CLASS__.'->'.__FUNCTION__.'->verify3D: '.$check3dId.print_r($res, TRUE));
      return $res;
    }

    /**
     * handleNO3DAction
     *
     * @param $post
     *
     * @return array
     */
    public function handleNO3DAction($post)
    {
      $returnURL = urldecode($post['returnURL']);
      return array('state' => 'paid', 'Location' => $returnURL);
    }

    /**
     * initCCardIFrame
     *
     * @param $cfg
     * @param $data
     * @param $translation
     *
     * @return string
     */
    public function initCCardIFrame($cfg, $data, $translation)
    {
      if (!empty($cfg['payformid'])) $this->payformid = $cfg['payformid'];
      if (empty($this->payformid)) $this->payformid = 'conditions-to-approve';

      $filename = dirname(__FILE__).'/../view/mcpay_form_ccard.tpl';
      $options  = array(
        'showRemember'     => TRUE,  // show remember checkbox
        'showRemove'       => FALSE, // show remove checkbox
        'cardStyle'        => TRUE,  // display in card style
        'futureYears'      => 10,
        'preSeletActDate'  => FALSE,
        'showMonthNames'   => FALSE,
        'specialBrand'     => FALSE,
        'specialHolder'    => FALSE,
        'preselectNewCard' => FALSE,
        'prefillPost'      => $_POST,
        'hideImg'          => TRUE,  // hide logo
        'imgStyle'         => '',     // logo style
      );
      if (!empty($_POST['mcpay_card_reuse']) && $_POST['mcpay_card_reuse'] == 'new') {
        $options['preselectNewCard'] = TRUE;
      }

      $accountMode = 0;
      if (!empty($cfg['accountMode'])) $accountMode = (int)$cfg['accountMode'];
      if ($accountMode == 1) $options['showRemember'] = FALSE;

      // mark error fields
      if (!empty($data['classes'])) $options['classes'] = $data['classes'];
      $errorMsg = '';
      if (!empty($data['errorMSG'])) $errorMsg = $data['errorMSG'];

      if (!empty($data['mcpay_fields'])) {
        $options['checkRemember'] = $data['mcpay_fields']['mcpay_card_remember'];
        $options['prefillPost']   = $data['mcpay_fields'];
      }

      $form = '';
      if (!empty($errorMsg)) {
        $form .= '<br><div class="'.$this->prefixCCard.'errorBox">'.$errorMsg.'</div>';
      }

      $prefillData = array();

      $scriptURL = '';
      if (!empty($cfg['jsScriptURL'])) {
        $scriptURL          = $cfg['jsScriptURL'];
        $this->plugin_urljs = $scriptURL;
      }
      $styleURL = '';
      if (!empty($cfg['jsStyleURL'])) {
        $styleURL            = $cfg['jsStyleURL'];
        $this->plugin_urlcss = $styleURL;
      }
      if (isset($cfg['fileExtJS'])) {
        $this->fileExtJS = $cfg['fileExtJS'];
      }
      $form .= $this->getCCardTPL($filename, $options, $prefillData, $translation);
      $form .= '<script>function mcpay_ccard_execute(){'
        .'var f = getElem("'.$this->payformid.'");'
        .'var s = document.createElement("script");'
        .'s.setAttribute( "src", "'.$scriptURL.'mcpay_formstyle_ccard_check" );'
        .'s.setAttribute( "type", "text/javascript");'
        .'f.appendChild(s);'
        .'return true;'
        .'}'
        .'getElem = function (name) { return document.getElementById(name); }</script>';

      $form .= '<script>';
      if ($this->debug) $form .= 'console.log("MCPAY ModelCCard START"); ';
      $inlineCode = $this->getStartJS(TRUE, FALSE, $this->debug, $this->payformid);
      $form       .= 'mcpay_ccard_execute(); ';
      $form       .= $inlineCode.' ';
      if ($this->debug) $form .= 'console.log("MCPAY ModelCCard END"); ';
      $form .= '</script>';
      $form .= '<link href="'.$styleURL.'mcpay_formstyle_ccard" rel="stylesheet" />';

      return $form;
    }

  } // end of class
}
?>
